//==========================================================================
//      					at91_mci.c
//							----------
//
// Low level driver for MMC/SD card using the MCI controller
//==========================================================================
// Author(s):   Ilko Iliev
// www.ronetix.at
// Contributors:
// Date:        16-03-2006
// Purpose:     
// Description: 
//==========================================================================
#include <stdlib.h>
#include <string.h>
#include <cyg/hal/hal_arch.h>
#include <pkgconf/devs_disk_at91_mci.h>
#include <cyg/io/user_config_keys.h>

#include <cyg/io/devtab.h>
#include <cyg/io/disk.h>
#include <cyg/hal/drv_api.h>
#include "at91_mci.h"

#define AT91C_MCI_TIMEOUT			1000000   				// For AT91F_MCIDeviceWaitReady 

static cyg_handle_t  				MCI_interrupt_handle;
static cyg_interrupt 				MCI_intrrupt;
static cyg_sem_t     				MCI_sem;

static AT91S_MciDeviceFeatures		MCI_Device_Features;
static AT91S_MciDeviceDesc			MCI_Device_Desc;
static MCI_CHAN						MCI_Device;

static cyg_mutex_t					MCI_mutex;


static int mmc_bread(disk_channel *priv, void *buf, cyg_uint32 len, cyg_uint32 block_num);
static int mmc_bwrite(disk_channel *priv, const void *buf, cyg_uint32 len, cyg_uint32 block_num);
static int mmc_get_config(disk_channel *priv, cyg_uint32 key, const void *buffer, cyg_uint32 *len);
static int mmc_set_config(disk_channel *priv, cyg_uint32 key, const void *buffer, cyg_uint32 *len);


static bool mmc_init(struct cyg_devtab_entry *tab);
static int  mmc_lookup(struct cyg_devtab_entry **tab, struct cyg_devtab_entry *st, const char *name);


static DISK_FUNS(mmc_disk_funs, 
                 mmc_bread, 
                 mmc_bwrite, 
                 mmc_get_config,
                 mmc_set_config
);


// ----------------------------------------------------------------------------

#define MMC_DISK_INSTANCE(_number_,_mbr_supp_,_name_)					\
static MCI_CHAN mmc_disk_info##_number_ = {                      		\
    .pMCI_DeviceDesc = &MCI_Device_Desc,								\
    .pMCI_DeviceFeatures = &MCI_Device_Features,						\
};                                                                    	\
DISK_CHANNEL(mmc_disk_channel##_number_,                               	\
             mmc_disk_funs,                                            	\
             mmc_disk_info##_number_,                                  	\
             _mbr_supp_,                                              	\
             4                                                        	\
);                                                                    	\
BLOCK_DEVTAB_ENTRY(mmc_disk_io##_number_,                              	\
                   _name_,                                            	\
                   0,                                                 	\
                   &cyg_io_disk_devio,                                	\
                   mmc_init,                                      		\
                   mmc_lookup,                                    		\
                   &mmc_disk_channel##_number_                         	\
);

MMC_DISK_INSTANCE(0, false, CYGDAT_DEV_DISK_AT91_MCI_NAME);			// without partition table	
MMC_DISK_INSTANCE(1, true,  CYGDAT_DEV_DISK_AT91_MCI_NAME"/");		// with partition table


// AT91RM9200 has a bug and needs swap
//====================================
#ifdef AT91_MCI_FLAG_SWAP
	#define	SWAP_ENDIAN_FUNC(buff,nn)		swap_endian(buff, nn)
#else
	#define	SWAP_ENDIAN_FUNC(buff,nn)
#endif

//=============================================================================
//							swap_endian()
//							-------------
//	Parameter:
//		buff   -> pointer to buffer 
//		ndword -> number of 32-bit words
//	
//	Return: none
//	Description:
//		swap big/little endian
//=============================================================================
static void swap_endian( void *buff, int ndwords )
{
int ii;
char *pp, tmp;

	pp = (char *)buff;
	for( ii=0; ii < ndwords; ii++ )        
	{
		tmp   = pp[0];
		pp[0] = pp[3];
		pp[3] = tmp;
		
		tmp   = pp[1];
		pp[1] = pp[2];
		pp[2] = tmp;
		
		pp += 4;		
	}
}


//*----------------------------------------------------------------------------
//* \fn    AT91F_CfgDevice
//* \brief This function is used to initialise MMC or SDCard Features
//*----------------------------------------------------------------------------
static void AT91F_CfgDevice(void)
{
	// Init Device Structure
	//======================
	MCI_Device_Features.Relative_Card_Address 		= 0;
	MCI_Device_Features.Card_Inserted 				= CARD_STATE_REMOVED;
	MCI_Device_Features.Max_Read_DataBlock_Length	= 0;
	MCI_Device_Features.Max_Write_DataBlock_Length 	= 0;
	MCI_Device_Features.Read_Partial 				= 0;
	MCI_Device_Features.Write_Partial 				= 0;
	MCI_Device_Features.Erase_Block_Enable 			= 0;
	MCI_Device_Features.Sector_Size 				= 0;
	MCI_Device_Features.Memory_Capacity 			= 0;
	MCI_Device_Desc.state							= AT91C_MCI_IDLE;
	MCI_Device_Desc.SDCard_bus_width				= AT91C_MCI_SCDBUS;

	MCI_Device.pMCI_DeviceDesc 		= &MCI_Device_Desc;
	MCI_Device.pMCI_DeviceFeatures 	= &MCI_Device_Features;
	MCI_Device.tx_buff				= NULL;
	MCI_Device.tx_buff_size			= 0;
}



//*----------------------------------------------------------------------------
//* \fn    AT91F_MCIDeviceWaitReady
//* \brief Wait for MCI Device ready
//*----------------------------------------------------------------------------
static int AT91F_MCIDeviceWaitReady(int timeout)
{
volatile int status;
	
	do
	{
		status = AT91C_BASE_MCI->MCI_SR;
		timeout--;
	}
	while( !(status & AT91C_MCI_NOTBUSY)  && (timeout > 0) );	
	
	return (timeout > 0) ? ENOERR : EIO;
}


//*----------------------------------------------------------------------------
//* \fn    AT91F_MCI_SendCommand
//* \brief Generic function to send a command to the MMC or SDCard
//*----------------------------------------------------------------------------
static int AT91F_MCI_SendCommand( MCI_CHAN *chan, cyg_uint32 Cmd, cyg_uint32 Arg)
{
volatile cyg_uint32 status;	
cyg_uint32	error;
int	ii=0;

    // Send the command
    //=================
    AT91C_BASE_MCI->MCI_ARGR = Arg;
    AT91C_BASE_MCI->MCI_CMDR = Cmd;

	// wait for CMDRDY Status flag to read the response
	//=================================================
	do
	{
		status = AT91C_BASE_MCI->MCI_SR;
	}
	while( !(status & AT91C_MCI_CMDRDY) && (++ii < AT91C_MCI_TIMEOUT) );
	
	if ( ii >= AT91C_MCI_TIMEOUT )
		return ETIMEDOUT;

    // Test error  ==> if crc error and response R3 ==> don't check error
    //===================================================================
    error = (AT91C_BASE_MCI->MCI_SR) & AT91C_MCI_SR_ERROR;
	if (error != 0 )
	{
		// if the command is SEND_OP_COND the CRC error flag is always present (cf : R3 response)
		//=======================================================================================
		if ( (Cmd != AT91C_SDCARD_APP_OP_COND_CMD) && (Cmd != AT91C_MMC_SEND_OP_COND_CMD) )
			return EIO;
		else
		{
			if (error != AT91C_MCI_RCRCE)
				return EIO;
		}
	}
    return ENOERR;
}

//*----------------------------------------------------------------------------
//* \fn    AT91F_MCI_SDCard_SendAppCommand
//* \brief Specific function to send a specific command to the SDCard
//*----------------------------------------------------------------------------
static int AT91F_MCI_SDCard_SendAppCommand(	MCI_CHAN *chan, cyg_uint32 Cmd_App,	cyg_uint32 Arg)
{
cyg_uint32 status;
int	ii=0;

	// Send the CMD55 for application specific command
	//================================================
    AT91C_BASE_MCI->MCI_ARGR = (chan->pMCI_DeviceFeatures->Relative_Card_Address << 16 );
    AT91C_BASE_MCI->MCI_CMDR = AT91C_APP_CMD;

	// wait for CMDRDY Status flag to read the response
	//=================================================
	do
	{
		status = AT91C_BASE_MCI->MCI_SR;
	}
	while( !(status & AT91C_MCI_CMDRDY) && (++ii < AT91C_MCI_TIMEOUT) );	
	
	if ( ii >= AT91C_MCI_TIMEOUT )
		return ETIMEDOUT;

    // if an error occurs
    //===================
    if (((AT91C_BASE_MCI->MCI_SR) & AT91C_MCI_SR_ERROR) != 0 )
		return EIO;

    // check if it is a specific command and then send the command
    //============================================================
	if ( (Cmd_App && AT91C_SDCARD_APP_ALL_CMD) == 0)
		return EIO;

   return( AT91F_MCI_SendCommand(chan, Cmd_App,Arg) );
}

//*----------------------------------------------------------------------------
//* \fn    AT91F_MCI_Get_Status
//* \brief Addressed card sends its status register
//*----------------------------------------------------------------------------
static int AT91F_MCI_Get_Status(MCI_CHAN *chan, cyg_uint32 relative_card_address, cyg_uint32 *card_stat)
{
int stat;

	stat = AT91F_MCI_SendCommand(chan, AT91C_SEND_STATUS_CMD, relative_card_address <<16);	
	if ( stat ) return stat;
	
	*card_stat = AT91C_BASE_MCI->MCI_RSPR[0];
	
   	return ENOERR;;
}


//*----------------------------------------------------------------------------
//* \fn    AT91F_MCI_ReadBlock
//* \brief Read an ENTIRE block or PARTIAL block
//*----------------------------------------------------------------------------
int AT91F_MCI_ReadBlock( MCI_CHAN *chan, int src, cyg_uint32 *dataBuffer, int sizeToRead )
{
int ii, stat;
cyg_uint32 rr;

    ////////////////////////////////////////////////////////////////////////////////////////////
    if(chan->pMCI_DeviceDesc->state != AT91C_MCI_IDLE)
    	return EIO;
    
    if ( (src + sizeToRead) > chan->pMCI_DeviceFeatures->Memory_Capacity )
		return EINVAL;

    // If source does not fit a begin of a block
    //==========================================
	if ( (src % chan->pMCI_DeviceFeatures->Max_Read_DataBlock_Length) != 0 )
		return EINVAL;
   
     // Test if the MMC supports Partial Read Block
     // ALWAYS SUPPORTED IN SD Memory Card
     //============================================
     if( (sizeToRead < chan->pMCI_DeviceFeatures->Max_Read_DataBlock_Length) 
    	&& (chan->pMCI_DeviceFeatures->Read_Partial == 0x00) )
   		return EINVAL;
   		
    if( sizeToRead > chan->pMCI_DeviceFeatures->Max_Read_DataBlock_Length)
   		return EINVAL;
    ////////////////////////////////////////////////////////////////////////////////////////////
	

	cyg_uint32 status;
	ii=0;
	do 
	{
		stat = AT91F_MCI_Get_Status(chan, chan->pMCI_DeviceFeatures->Relative_Card_Address, &status);
		if ( stat ) return stat;
	}
	while ( !(status & AT91C_SR_READY_FOR_DATA) && ++ii < AT91C_MCI_TIMEOUT);
	
	if ( ii >= AT91C_MCI_TIMEOUT )
		return ETIMEDOUT;
	
    // Init Mode Register
    //===================
    rr = AT91C_BASE_MCI->MCI_MR;
    rr &= 0x7FFF;
	AT91C_BASE_MCI->MCI_MR = rr | ((chan->pMCI_DeviceFeatures->Max_Read_DataBlock_Length << 16) | AT91C_MCI_PDCMODE);

	DCACHE_FLUSH_RANGE( dataBuffer, sizeToRead );
	 
    if (sizeToRead %4)
		sizeToRead = (sizeToRead /4)+1;
	else
		sizeToRead = sizeToRead/4;

	AT91C_BASE_PDC_MCI->PDC_PTCR = (AT91C_PDC_TXTDIS | AT91C_PDC_RXTDIS);
    AT91C_BASE_PDC_MCI->PDC_RPR  = (cyg_uint32)dataBuffer;
    AT91C_BASE_PDC_MCI->PDC_RCR  = sizeToRead;

	chan->pMCI_DeviceDesc->state = AT91C_MCI_RX_SINGLE_BLOCK;

    AT91C_BASE_MCI->MCI_IER = AT91C_MCI_RXBUFF;

	// (PDC) Receiver Transfer Enable
	//===============================
	AT91C_BASE_PDC_MCI->PDC_PTCR = AT91C_PDC_RXTEN;
	
	// Send the Read single block command
	//===================================
	stat = AT91F_MCI_SendCommand(chan, AT91C_READ_SINGLE_BLOCK_CMD, src);
    if ( stat ) return stat; 

    stat = cyg_semaphore_timed_wait( &MCI_sem, cyg_current_time() + 50 );
    if ( stat == false )
    {
        stat =  ETIMEDOUT;
		chan->pMCI_DeviceDesc->state = AT91C_MCI_IDLE;        
	}
    else
        stat = ENOERR;

	// there is a bug in AT91RM9200 - the buffer must be endian swapped
	//==================================================================
	SWAP_ENDIAN_FUNC( dataBuffer, sizeToRead );        
	
	return stat;
}

//*----------------------------------------------------------------------------
//* \fn    AT91F_MCI_WriteBlock
//* \brief  Write an ENTIRE block but not always PARTIAL block !!!
//*----------------------------------------------------------------------------
int AT91F_MCI_WriteBlock( MCI_CHAN *chan, int dest, cyg_uint32 *dataBuffer, int sizeToWrite )
{
int stat, ii=0;    
cyg_uint32 status, rr;

	if( chan->pMCI_DeviceDesc->state != AT91C_MCI_IDLE)
    	return EIO;
    	
    if ( (dest + sizeToWrite) > chan->pMCI_DeviceFeatures->Memory_Capacity )
		return EINVAL;

    // If source does not fit a begin of a block
    //==========================================
	if ( (dest % chan->pMCI_DeviceFeatures->Max_Read_DataBlock_Length) != 0 )
		return EINVAL;
   
    // Test if the MMC supports Partial Write Block 
    //=============================================
    if( (sizeToWrite < chan->pMCI_DeviceFeatures->Max_Write_DataBlock_Length) 
    	&& (chan->pMCI_DeviceFeatures->Write_Partial == 0x00) )
   		return EINVAL;
   		
   	if( sizeToWrite > chan->pMCI_DeviceFeatures->Max_Write_DataBlock_Length )
   		return EINVAL;
    ////////////////////////////////////////////////////////////////////////////////////////////

	do 
	{
		stat = AT91F_MCI_Get_Status(chan, chan->pMCI_DeviceFeatures->Relative_Card_Address, &status);
		if ( stat ) return stat;
		
	} while ( !(status & AT91C_SR_READY_FOR_DATA) && ++ii < AT91C_MCI_TIMEOUT );

	if ( ii >= AT91C_MCI_TIMEOUT )
		return ETIMEDOUT;

	DCACHE_FLUSH_RANGE( chan->tx_buff, sizeToWrite );

	if (sizeToWrite %4)
		sizeToWrite = (sizeToWrite /4)+1;
	else
		sizeToWrite = sizeToWrite/4;

	// there is a bug in AT91RM9200 - the buffer must be endian swapped
	//==================================================================
	#ifdef SWAP_ENDIAN_FUNC
		memcpy( chan->tx_buff, dataBuffer, sizeToWrite*4 );
		SWAP_ENDIAN_FUNC( chan->tx_buff, sizeToWrite );  
	#endif	

  
    // Init Mode Register
    //===================
    rr = AT91C_BASE_MCI->MCI_MR;
    rr &= 0x7FFF;
	AT91C_BASE_MCI->MCI_MR = rr | ((chan->pMCI_DeviceFeatures->Max_Write_DataBlock_Length << 16) | AT91C_MCI_PDCMODE);
	
	// Init PDC for write sequence
	//============================
    AT91C_BASE_PDC_MCI->PDC_PTCR = (AT91C_PDC_TXTDIS | AT91C_PDC_RXTDIS);
    AT91C_BASE_PDC_MCI->PDC_TPR = (cyg_uint32) chan->tx_buff;
    AT91C_BASE_PDC_MCI->PDC_TCR = sizeToWrite;

	chan->pMCI_DeviceDesc->state = AT91C_MCI_TX_SINGLE_BLOCK;

	// Enable AT91C_MCI_TXBUFE Interrupt
	//==================================
    AT91C_BASE_MCI->MCI_IER = AT91C_MCI_TXBUFE;

	// Send the write single block command
	//====================================
	stat = AT91F_MCI_SendCommand(chan, AT91C_WRITE_BLOCK_CMD, dest);
	if ( stat ) return stat;


  	// Enables TX for PDC transfert requests
  	//======================================
    AT91C_BASE_PDC_MCI->PDC_PTCR = AT91C_PDC_TXTEN;

    stat = cyg_semaphore_timed_wait( &MCI_sem, cyg_current_time() + 250 );
    if ( stat == false )
    {
        stat =  ETIMEDOUT;
        chan->pMCI_DeviceDesc->state = AT91C_MCI_IDLE;
    }
    else
        stat = ENOERR;

	return stat;
}

//*------------------------------------------------------------------------------------------------------------
//* \fn    AT91F_MCI_MMC_SelectCard
//* \brief Toggles a card between the Stand_by and Transfer states or between Programming and Disconnect states
//*------------------------------------------------------------------------------------------------------------
static int AT91F_MCI_MMC_SelectCard( MCI_CHAN *chan, cyg_uint32 relative_card_address)
{
cyg_uint32 status;
int stat, ii;
	
	// Check if the MMC card chosen is already the selected one
	//=========================================================
	stat = AT91F_MCI_Get_Status(chan,relative_card_address, &status);
	if (stat) return stat;

	if ((status & AT91C_SR_CARD_SELECTED) == AT91C_SR_CARD_SELECTED)
		return ENOERR;

	// Search for the MMC Card to be selected, ii = the Corresponding Device Number
	//=================================================================================
	ii = 0;
	while( (chan->pMCI_DeviceFeatures[ii].Relative_Card_Address != relative_card_address)
		   && (status < AT91C_MAX_MCI_CARDS) )
		status++;

	if (ii > AT91C_MAX_MCI_CARDS)
    	return EINVAL;

	stat = AT91F_MCI_SendCommand( chan, AT91C_SEL_DESEL_CARD_CMD, chan->pMCI_DeviceFeatures[status].Relative_Card_Address << 16);
	if ( stat ) return stat;

    return ENOERR;
}

//*----------------------------------------------------------------------------
//* \fn    AT91F_MCI_GetCSD
//* \brief Asks to the specified card to send its CSD
//*----------------------------------------------------------------------------
static int AT91F_MCI_GetCSD (MCI_CHAN *chan, cyg_uint32 relative_card_address , cyg_uint32 * response)
{
int stat;

	stat = AT91F_MCI_SendCommand(chan, AT91C_SEND_CSD_CMD, (relative_card_address << 16)); 	
	if ( stat ) return stat;
	
    response[0] = AT91C_BASE_MCI->MCI_RSPR[0];
   	response[1] = AT91C_BASE_MCI->MCI_RSPR[1];
    response[2] = AT91C_BASE_MCI->MCI_RSPR[2];
    response[3] = AT91C_BASE_MCI->MCI_RSPR[3];
    
    return ENOERR;
}

//*----------------------------------------------------------------------------
//* \fn    AT91F_MCI_SetBlocklength
//* \brief Select a block length for all following block commands (R/W)
//*----------------------------------------------------------------------------
static int AT91F_MCI_SetBlocklength( MCI_CHAN *chan, cyg_uint32 length)
{
    return( AT91F_MCI_SendCommand(chan, AT91C_SET_BLOCKLEN_CMD, length) );
}

//*----------------------------------------------------------------------------
//* \fn    AT91F_MCI_MMC_GetAllOCR
//* \brief Asks to all cards to send their operations conditions
//*----------------------------------------------------------------------------
static int AT91F_MCI_MMC_GetAllOCR( MCI_CHAN *chan )
{
cyg_uint32	resp;
int stat, ii=0;

 	do
    {
    	stat = AT91F_MCI_SendCommand( chan, AT91C_MMC_SEND_OP_COND_CMD, AT91C_MMC_HOST_VOLTAGE_RANGE);
    	if ( stat ) return stat;
		
		resp  = AT91C_BASE_MCI->MCI_RSPR[0];
		resp &= AT91C_CARD_POWER_UP_BUSY;
		
	}
	while( resp != AT91C_CARD_POWER_UP_BUSY && ++ii < AT91C_MCI_TIMEOUT );
	
	if ( ii >= AT91C_MCI_TIMEOUT )
		return ETIMEDOUT;
	
	return ENOERR;
}

//*----------------------------------------------------------------------------
//* \fn    AT91F_MCI_MMC_GetAllCID
//* \brief Asks to the MMC on the chosen slot to send its CID
//*----------------------------------------------------------------------------
static int AT91F_MCI_MMC_GetAllCID( MCI_CHAN *chan, cyg_uint32 *response, int *num_card)
{
int stat;	
int Nb_Cards_Found=-1;
  
  	*num_card = 0;
	while(1)
	{
		stat = AT91F_MCI_SendCommand(chan, AT91C_MMC_ALL_SEND_CID_CMD, AT91C_NO_ARGUMENT);
		if ( stat ) break;
		
		Nb_Cards_Found = 0;
		
		// Assignation of the relative address to the MMC CARD
		//====================================================
		chan->pMCI_DeviceFeatures[Nb_Cards_Found].Relative_Card_Address = Nb_Cards_Found + AT91C_FIRST_RCA;
		
		// Set the insert flag
		//====================
		chan->pMCI_DeviceFeatures[Nb_Cards_Found].Card_Inserted = CARD_STATE_MMC_INSERTED;

		stat = AT91F_MCI_SendCommand(chan, AT91C_MMC_SET_RELATIVE_ADDR_CMD, (Nb_Cards_Found + AT91C_FIRST_RCA) << 16);
		if ( stat ) return stat;								 
								 
		// If no error during assignation address ==> Increment Nb_cards_Found
		//====================================================================
		Nb_Cards_Found++ ;
		(*num_card)++;
	}
	
	return ENOERR;
}

//*----------------------------------------------------------------------------
//* \fn    AT91F_MCI_MMC_Init
//* \brief Return the MMC initialisation status
//*----------------------------------------------------------------------------
static int AT91F_MCI_MMC_Init( MCI_CHAN *chan )
{
cyg_uint32	tab_response[4];
cyg_uint32	mult, blocknr;
cyg_uint32 	i, Nb_Cards_Found=0;
int csd_ver, stat;

	AT91F_MCI_Configure(AT91C_BASE_MCI,
						AT91C_MCI_DTOR_1MEGA_CYCLES,
						0x834A,								// 400kHz for MCK = 60MHz
						AT91C_MCI_MMC_SLOTA);

	// Wait MCI Device Ready
	//======================
	stat = AT91F_MCIDeviceWaitReady(AT91C_MCI_TIMEOUT);
	if ( stat ) return stat;

	// Resets all MMC Cards in Idle state
	//===================================
	AT91F_MCI_SendCommand(chan, AT91C_MMC_GO_IDLE_STATE_CMD, AT91C_NO_ARGUMENT);

	stat = AT91F_MCI_MMC_GetAllOCR(chan);
	if (stat) return stat;
	

	stat = AT91F_MCI_MMC_GetAllCID(chan,tab_response, &Nb_Cards_Found);
	if ( stat ) return stat;
	
	    // Set the Mode Register
	    //======================
    	AT91C_BASE_MCI->MCI_MR = AT91C_MCI_PDCMODE;

		for(i = 0; i < Nb_Cards_Found; i++)
		{
			stat = AT91F_MCI_GetCSD(chan, chan->pMCI_DeviceFeatures[i].Relative_Card_Address, tab_response);
			if ( stat ) 
				chan->pMCI_DeviceFeatures[i].Relative_Card_Address = 0;					  
			else
			{
				csd_ver = (tab_response[3] >> AT91C_CSD_STRUCT_S) & AT91C_CSD_STRUCT_M;
				if ( csd_ver !=1 && csd_ver != 2 )
					return EIO;
				
				chan->pMCI_DeviceFeatures[i].Max_Read_DataBlock_Length 	= 1 << ((tab_response[1] >> AT91C_CSD_RD_B_LEN_S) & AT91C_CSD_RD_B_LEN_M );
	 			chan->pMCI_DeviceFeatures[i].Max_Write_DataBlock_Length 	=	1 << ((tab_response[3] >> AT91C_CSD_WBLEN_S) & AT91C_CSD_WBLEN_M );
				chan->pMCI_DeviceFeatures[i].Sector_Size 				= 1 + ((tab_response[2] >> AT91C_CSD_v22_SECT_SIZE_S) & AT91C_CSD_v22_SECT_SIZE_M );
		  		chan->pMCI_DeviceFeatures[i].Read_Partial 				= (tab_response[1] >> AT91C_CSD_RD_B_PAR_S) & AT91C_CSD_RD_B_PAR_M;
				chan->pMCI_DeviceFeatures[i].Write_Partial 				= (tab_response[3] >> AT91C_CSD_WBLOCK_P_S) & AT91C_CSD_WBLOCK_P_M;
				
				// None in MMC specification version 2.2
				//======================================
				chan->pMCI_DeviceFeatures[i].Erase_Block_Enable = 0;
				
				chan->pMCI_DeviceFeatures[i].Read_Block_Misalignment = (tab_response[1] >> AT91C_CSD_RD_B_MIS_S) & AT91C_CSD_RD_B_MIS_M;
				chan->pMCI_DeviceFeatures[i].Write_Block_Misalignment = (tab_response[1] >> AT91C_CSD_WR_B_MIS_S) & AT91C_CSD_WR_B_MIS_M;

				// Compute Memory Capacity
				//========================
				mult = 1 << ( ((tab_response[2] >> AT91C_CSD_C_SIZE_M_S) & AT91C_CSD_C_SIZE_M_M) + 2 );				// compute MULT
				
				blocknr = ((tab_response[1] >> AT91C_CSD_CSIZE_H_S) & AT91C_CSD_CSIZE_H_M) << 2;					// compute MSB of C_SIZE

				// compute MULT * (LSB of C-SIZE + MSB already computed + 1) = BLOCKNR
				//====================================================================
				blocknr = mult * ( ( blocknr + ( (tab_response[2] >> AT91C_CSD_CSIZE_L_S) & AT91C_CSD_CSIZE_L_M) ) + 1 );
				chan->pMCI_DeviceFeatures[i].num_pages = blocknr;

				chan->pMCI_DeviceFeatures[i].Memory_Capacity =  chan->pMCI_DeviceFeatures[i].Max_Read_DataBlock_Length * blocknr;

		  		
			}
		}	// end of for
		

	return ENOERR;
}

//*----------------------------------------------------------------------------
//* \fn    AT91F_MCI_SDCard_GetOCR
//* \brief Asks to all cards to send their operations conditions
//*----------------------------------------------------------------------------
static int AT91F_MCI_SDCard_GetOCR( MCI_CHAN *chan, cyg_uint32 *card_stat )
{
cyg_uint32 resp;
int stat, ii=0;

	// The RCA to be used for CMD55 in Idle state shall be the card's default RCA=0x0000.
	//===================================================================================
	chan->pMCI_DeviceFeatures->Relative_Card_Address = 0x0;

	do
	{
    	stat = AT91F_MCI_SDCard_SendAppCommand(chan, AT91C_SDCARD_APP_OP_COND_CMD, AT91C_MMC_HOST_VOLTAGE_RANGE);
		if (stat) return stat;
		
		resp = AT91C_BASE_MCI->MCI_RSPR[0];
		resp &= AT91C_CARD_POWER_UP_BUSY;
	}
	while ( resp != AT91C_CARD_POWER_UP_BUSY && ++ii < AT91C_MCI_TIMEOUT );
	
	if ( ii >= AT91C_MCI_TIMEOUT )
		return ETIMEDOUT;

	*card_stat = AT91C_BASE_MCI->MCI_RSPR[0];
	
	return ENOERR;
}

//*----------------------------------------------------------------------------
//* \fn    AT91F_MCI_SDCard_GetCID
//* \brief Asks to the SDCard on the chosen slot to send its CID
//*----------------------------------------------------------------------------
static int AT91F_MCI_SDCard_GetCID (MCI_CHAN *chan, cyg_uint32 *response)
{
int stat;

	stat = AT91F_MCI_SendCommand(chan, AT91C_ALL_SEND_CID_CMD, AT91C_NO_ARGUMENT);
	if ( stat ) return stat;
	
    response[0] = AT91C_BASE_MCI->MCI_RSPR[0];
   	response[1] = AT91C_BASE_MCI->MCI_RSPR[1];
    response[2] = AT91C_BASE_MCI->MCI_RSPR[2];
    response[3] = AT91C_BASE_MCI->MCI_RSPR[3];
    
    return ENOERR;
}

//*----------------------------------------------------------------------------
//* \fn    AT91F_MCI_SDCard_SetBusWidth
//* \brief  Set bus width for SDCard
//*----------------------------------------------------------------------------
static int AT91F_MCI_SDCard_SetBusWidth(MCI_CHAN *chan)
{
cyg_uint32	ret_value;
char bus_width;
int stat, ii=0;

	do
	{
		stat = AT91F_MCI_Get_Status(chan,chan->pMCI_DeviceFeatures->Relative_Card_Address, &ret_value);
		if ( stat ) return stat;
	}
	while( ((ret_value & AT91C_SR_READY_FOR_DATA) == 0) && ++ii < AT91C_MCI_TIMEOUT );
	
	if ( ii >= AT91C_MCI_TIMEOUT )
		return ETIMEDOUT;

	// Select Card
	//============
    AT91F_MCI_SendCommand(chan, AT91C_SEL_DESEL_CARD_CMD, (chan->pMCI_DeviceFeatures->Relative_Card_Address)<<16);

	// Set bus width for SD card
	//==========================
	if(chan->pMCI_DeviceDesc->SDCard_bus_width == AT91C_MCI_SCDBUS)
		 	bus_width = AT91C_BUS_WIDTH_4BITS;
	else	bus_width = AT91C_BUS_WIDTH_1BIT;


	stat = AT91F_MCI_SDCard_SendAppCommand(chan,AT91C_SDCARD_SET_BUS_WIDTH_CMD,bus_width);
	if ( stat ) return stat;
	
	return ENOERR;
}

//*----------------------------------------------------------------------------
//* \fn    AT91F_MCI_SDCard_Init
//* \brief Return the SDCard initialisation status
//*----------------------------------------------------------------------------
static int AT91F_MCI_SDCard_Init( MCI_CHAN *chan )
{
cyg_uint32	tab_response[4];
cyg_uint32	mult,blocknr;
int csd_ver, stat;

	AT91F_MCI_SW_Reset( AT91C_BASE_MCI );

	AT91F_MCI_Configure(AT91C_BASE_MCI,
						AT91C_MCI_DTOR_1MEGA_CYCLES,
						AT91C_MCI_PDCMODE,				
						AT91C_MCI_SDCARD_4BITS_SLOTA);

	AT91F_MCI_CfgFrequency( AT91C_BASE_MCI, CYGNUM_HAL_ARM_AT91_CLOCK_SPEED, CYGDAT_DEV_DISK_AT91_MCI_CLK*1000000 );

	AT91F_MCI_SendCommand(chan, AT91C_GO_IDLE_STATE_CMD, AT91C_NO_ARGUMENT);
	AT91F_MCI_SendCommand(chan, AT91C_GO_IDLE_STATE_CMD, AT91C_NO_ARGUMENT);

	stat = AT91F_MCI_SDCard_GetOCR(chan, &tab_response[0] );
    if (stat) return stat;


	stat = AT91F_MCI_SDCard_GetCID(chan, tab_response);
	if ( stat ) return stat;

    chan->pMCI_DeviceFeatures->Card_Inserted = CARD_STATE_SD_INSERTED;

	stat = AT91F_MCI_SendCommand(chan, AT91C_SET_RELATIVE_ADDR_CMD, 0);
	if (stat) return stat;
	
	chan->pMCI_DeviceFeatures->Relative_Card_Address = (AT91C_BASE_MCI->MCI_RSPR[0] >> 16);
	
	stat = AT91F_MCI_GetCSD(chan,chan->pMCI_DeviceFeatures->Relative_Card_Address,tab_response);
	if ( stat ) return stat;
	
	csd_ver = (tab_response[3] >> AT91C_CSD_STRUCT_S) & AT91C_CSD_STRUCT_M;
	if ( csd_ver < 0 || csd_ver > 2 ) return EIO;
				
	chan->pMCI_DeviceFeatures->Max_Read_DataBlock_Length 	= 1 << ((tab_response[1] >> AT91C_CSD_RD_B_LEN_S) & AT91C_CSD_RD_B_LEN_M );
	chan->pMCI_DeviceFeatures->Max_Write_DataBlock_Length 	= 1 << ((tab_response[3] >> AT91C_CSD_WBLEN_S) & AT91C_CSD_WBLEN_M );
	chan->pMCI_DeviceFeatures->Sector_Size 					= 1 + ((tab_response[2] >> AT91C_CSD_v21_SECT_SIZE_S) & AT91C_CSD_v21_SECT_SIZE_M );
	chan->pMCI_DeviceFeatures->Read_Partial 				= (tab_response[1] >> AT91C_CSD_RD_B_PAR_S) & AT91C_CSD_RD_B_PAR_M;
	chan->pMCI_DeviceFeatures->Write_Partial 				= (tab_response[3] >> AT91C_CSD_WBLOCK_P_S) & AT91C_CSD_WBLOCK_P_M;
	chan->pMCI_DeviceFeatures->Erase_Block_Enable 			= (tab_response[3] >> AT91C_CSD_v21_ER_BLEN_EN_S) & AT91C_CSD_v21_ER_BLEN_EN_M;
	chan->pMCI_DeviceFeatures->Read_Block_Misalignment 		= (tab_response[1] >> AT91C_CSD_RD_B_MIS_S) & AT91C_CSD_RD_B_MIS_M;
	chan->pMCI_DeviceFeatures->Write_Block_Misalignment		= (tab_response[1] >> AT91C_CSD_WR_B_MIS_S) & AT91C_CSD_WR_B_MIS_M;

	// Compute Memory Capacity
	//========================
	mult = 1 << ( ((tab_response[2] >> AT91C_CSD_C_SIZE_M_S) & AT91C_CSD_C_SIZE_M_M) + 2 );			// compute MULT
	blocknr = ((tab_response[1] >> AT91C_CSD_CSIZE_H_S) & AT91C_CSD_CSIZE_H_M) << 2;				// compute MSB of C_SIZE

	// compute MULT * (LSB of C-SIZE + MSB already computed + 1) = BLOCKNR
	//====================================================================
	blocknr = mult * ( ( blocknr + ( (tab_response[2] >> AT91C_CSD_CSIZE_L_S) & AT91C_CSD_CSIZE_L_M) ) + 1 );

	chan->pMCI_DeviceFeatures->num_pages = blocknr;

	chan->pMCI_DeviceFeatures->Memory_Capacity =  chan->pMCI_DeviceFeatures->Max_Read_DataBlock_Length * blocknr;

	stat = AT91F_MCI_SDCard_SetBusWidth(chan);
	if ( stat ) return stat;

	stat = AT91F_MCI_SetBlocklength(chan,chan->pMCI_DeviceFeatures->Max_Read_DataBlock_Length);
	if ( stat ) return stat;
				
    return ENOERR;
}



//=============================================================================
//								mci_ISR()
//								---------
//
//=============================================================================
static cyg_uint32 mci_ISR(cyg_vector_t vector, cyg_addrword_t data)
{
cyg_uint32 stat;
MCI_CHAN *chan = (MCI_CHAN *)data;

	stat =  AT91C_BASE_MCI->MCI_SR & AT91C_BASE_MCI->MCI_IMR;

	// If End of Tx Buffer Empty interrupt occurred
	//=============================================
	if ( stat & AT91C_MCI_TXBUFE )
    {
		AT91C_BASE_MCI->MCI_IDR = AT91C_MCI_TXBUFE;
 		AT91C_BASE_PDC_MCI->PDC_PTCR = AT91C_PDC_TXTDIS;
        	
		chan->pMCI_DeviceDesc->state = AT91C_MCI_IDLE;
	}	
	

    // If End of Rx Buffer Full interrupt occurred
    //============================================
    if ( stat & AT91C_MCI_RXBUFF )
    {        
       	AT91C_BASE_MCI->MCI_IDR = AT91C_MCI_RXBUFF;
 		AT91C_BASE_PDC_MCI->PDC_PTCR = AT91C_PDC_RXTDIS;
	
		chan->pMCI_DeviceDesc->state = AT91C_MCI_IDLE;
	}	

    
//    cyg_drv_interrupt_mask(vector);
    cyg_drv_interrupt_acknowledge(vector);

    return CYG_ISR_CALL_DSR;
}

//=============================================================================
//								mci_DSR()
//								---------
//
//=============================================================================
static void mci_DSR(cyg_vector_t vector, cyg_ucount32 count, cyg_addrword_t data)
{
MCI_CHAN *chan = (MCI_CHAN *)data;

	chan->pMCI_DeviceDesc->state = AT91C_MCI_IDLE;
	cyg_semaphore_post( &MCI_sem);
}


//=====================================================================
//                      	mmc_init()
//                      	----------
//
//=====================================================================
static bool mmc_init(struct cyg_devtab_entry *tab)
{
MCI_CHAN *chan;

    chan = (MCI_CHAN *)tab->priv;
    
	cyg_mutex_init( &MCI_mutex );
	cyg_semaphore_init( &MCI_sem, 0);

	AT91F_MCI_SW_Reset( AT91C_BASE_MCI );

	// Init MCI for MMC and SDCard interface
	//======================================
	AT91F_MCI_CfgPIO();	
	AT91F_MCI_CfgPMC();
	AT91F_PDC_Open(AT91C_BASE_PDC_MCI);

    // Disable all the interrupts
    //===========================
    AT91C_BASE_MCI->MCI_IDR = 0xFFFFFFFF;
    
	// Init MCI Device Structures
	//===========================
	AT91F_CfgDevice();

    // Create and attach SPI interrupt object
    //=======================================
    cyg_drv_interrupt_create(AT91C_ID_MCI,
                             4,                   
                             (cyg_addrword_t)&MCI_Device,   					// data to ISR and DSR
                             mci_ISR,
                             mci_DSR,
                             &MCI_interrupt_handle,
                             &MCI_intrrupt);

    cyg_drv_interrupt_attach(MCI_interrupt_handle);
    cyg_drv_interrupt_unmask(AT91C_ID_MCI);
    
    return ENOERR;
}


//=====================================================================
//                      	mmc_lookup()
//                      	------------
//  Parameter:
//		tab  ->
//		st   ->
//		name ->
//
//  Return:
//      EINVAL - if invalid port number or the port is already open
//
//  Description:
//      
//=====================================================================
static int mmc_lookup(struct cyg_devtab_entry **tab, struct cyg_devtab_entry *st, const char *name)
{
disk_channel *disk_chan = (disk_channel *) (*tab)->priv;	
MCI_CHAN *chan = (MCI_CHAN *)disk_chan->dev_priv;
int stat;

//    cyg_mutex_lock( &MCI_mutex );

	stat = EIO;
	while(1)
	{
		stat = AT91F_MCI_SDCard_Init(chan);
		if ( stat ) break;
	
		if ( chan->pMCI_DeviceFeatures->Max_Write_DataBlock_Length > MCI_MAX_BLOCK_SIZE ) break;
	
		if ( chan->pMCI_DeviceFeatures->Max_Write_DataBlock_Length != chan->tx_buff_size )
		{
			if ( chan->tx_buff_size > 0 && chan->tx_buff != NULL)
				free( chan->tx_buff );
				
			chan->tx_buff_size = chan->pMCI_DeviceFeatures->Max_Write_DataBlock_Length;	
			chan->tx_buff = malloc( chan->tx_buff_size );
			if ( chan->tx_buff == NULL )
				break;	
		}
		
		stat = ENOERR;
		break;
	}


	if ( stat != ENOERR )
	{
		stat = AT91F_MCI_MMC_Init(&MCI_Device);
		if (stat) return stat;
	
		if ( chan->pMCI_DeviceFeatures->Max_Write_DataBlock_Length != chan->tx_buff_size )
		{
			if ( chan->tx_buff_size > 0 && chan->tx_buff != NULL)
				free( chan->tx_buff );
				
			chan->tx_buff_size = chan->pMCI_DeviceFeatures->Max_Write_DataBlock_Length;	
			chan->tx_buff = malloc( chan->tx_buff_size );
			if ( chan->tx_buff == NULL )
				return EIO;
		}
	
		// Select MMC Card n1
		//====================
		stat = AT91F_MCI_MMC_SelectCard(&MCI_Device,AT91C_FIRST_RCA);
		if (stat) return stat;
	
		// Set Block Length
		//=================
		stat = AT91F_MCI_SetBlocklength(&MCI_Device,MCI_Device.pMCI_DeviceFeatures[0].Max_Read_DataBlock_Length);
		if (stat) return stat;
	}

//    cyg_mutex_unlock( &MCI_mutex );

    cyg_disk_identify_t ident;

	ident.serial[0]       = '\0';
	ident.firmware_rev[0] = '\0';
	ident.model_num[0]    = '\0';

    ident.lba_sectors_num = chan->pMCI_DeviceFeatures->Memory_Capacity / chan->pMCI_DeviceFeatures->Max_Write_DataBlock_Length;
    ident.cylinders_num   = 0;
    ident.heads_num       = 0;
    ident.sectors_num     = chan->pMCI_DeviceFeatures->Max_Write_DataBlock_Length;

    disk_chan->callbacks->disk_init(*tab);
        
    stat = disk_chan->callbacks->disk_connected(*tab, &ident);
    if ( stat ) return stat;    

    stat = disk_chan->callbacks->disk_lookup(tab, st, name);
	if ( stat ) return stat;

	return ENOERR;
}

//===================================================================================================
//                              mmc_bwrite()
//                              ------------
//  Paramater:
//      handle -> channel handle
//      buff   -> data to write
//      len    -> number of blocks to write
//		pos    -> start number of block
//
//  Return:
//      ENOERR -> if no error
//      EINVAL -> if invalid parameter
//===================================================================================================
static Cyg_ErrNo mmc_bwrite(disk_channel *priv, const void *buff, cyg_uint32 len, cyg_uint32 block_num)
{
MCI_CHAN *chan;
int stat, block_size;

    chan = (MCI_CHAN *)priv->dev_priv;
    
    block_size = chan->pMCI_DeviceFeatures->Max_Write_DataBlock_Length;
	
    cyg_mutex_lock( &MCI_mutex );

	stat = AT91F_MCI_WriteBlock( chan, block_num*block_size, (cyg_uint32 *)buff, len );
	
    cyg_mutex_unlock( &MCI_mutex );

    return stat;
}

//===================================================================================================
//                              mmc_bread()
//                              -----------
//  Paramater:
//      handle -> channel handle
//      buff   -> pointer to array where to store read data
//      len    -> number of blocks to read
//		pos    -> number of block to read
//
//  Return:
//      ENOERR -> if no error
//      EINVAL -> if invalid parameter
//===================================================================================================
static int mmc_bread(disk_channel *priv, void *buff, cyg_uint32 len, cyg_uint32 block_num)
{
MCI_CHAN *chan;
int stat, block_size;


	chan = (MCI_CHAN *)priv->dev_priv;

	block_size = chan->pMCI_DeviceFeatures->Max_Read_DataBlock_Length;

    cyg_mutex_lock( &MCI_mutex );
	
	stat = AT91F_MCI_ReadBlock( chan, block_num*block_size, (cyg_uint32 *)buff, len );

    cyg_mutex_unlock( &MCI_mutex );

    return stat;
}



//===================================================================================================
//                              mmc_set_config()
//                              ----------------
//  Parameter:
//      handle -> channel handle
//      key    -> CYG_IO_CONFIG_MMC_DIVIDER
//      buffer -> pointer to configuration data
//      len    -> pointer to variable that contain the size of buffer. MUST be exactly the size of
//                parameter's structure
//
//  Rerturn:
//      ENOERR - if no error
//      EINVAL - if the size of buffer don't mach the size of parameter's structure
//
//  Description:
//      Set diverse config parameters
//===================================================================================================
static int mmc_set_config(disk_channel *priv, cyg_uint32 key, const void *buffer, cyg_uint32 *len)
{
MCI_CHAN *chan;

    chan = (MCI_CHAN *)priv->dev_priv;

	switch ( key )
	{
		default:
			return ENOSYS;
			break;		
	}		

	
    return ENOERR;
}

//===================================================================================================
//                              mmc_get_config()
//                              ----------------
//  Parameter:
//      handle -> channel handle
//      key    -> CYG_IO_CONFIG_JTAG_DIVIDER
//      buffer -> pointer to configuration data
//      len    -> pointer to variable that contain the size of buffer. MUST be exactly the size of
//                parameter's structure
//
//  Rerturn:
//      ENOERR - if no error
//      EINVAL - if the size of buffer don't mach the size of parameter's structure
//
//  Description:
//      Return diverse config parameters
//===================================================================================================
static int mmc_get_config(disk_channel *priv, cyg_uint32 key, const void *buffer, cyg_uint32 *len)
{
MCI_CHAN *chan;

    chan = (MCI_CHAN *)priv->dev_priv;

	if ( (*len) != sizeof(cyg_uint32) )
		return EINVAL;

	switch ( key )
	{
		case CYG_IO_CONFIG_MMC_GET_RDBLOCK_SIZE:
			*(cyg_uint32 *)buffer = chan->pMCI_DeviceFeatures->Max_Read_DataBlock_Length;
			break;
	
		case CYG_IO_CONFIG_MMC_GET_WRBLOCK_SIZE:
			*(cyg_uint32 *)buffer = chan->pMCI_DeviceFeatures->Max_Write_DataBlock_Length;
			break;
	
		case CYG_IO_CONFIG_MMC_GET_SECTOR_SIZE:
			*(cyg_uint32 *)buffer = chan->pMCI_DeviceFeatures->Sector_Size;
			break;
			
		case CYG_IO_CONFIG_MMC_GET_CARD_SIZE:
			*(cyg_uint32 *)buffer = chan->pMCI_DeviceFeatures->Memory_Capacity;
			break;
			
		case CYG_IO_CONFIG_MMC_GET_NUM_PAGES:
			*(cyg_uint32 *)buffer = chan->pMCI_DeviceFeatures->num_pages;
			break;
			
	
		default:
			return ENOSYS;
			break;		
	}		

    return ENOERR;
}


