//======================================================================================
//
//                               at45_data_flash.c
//                               -----------------
// Device driver for Atmels's AT45 DataFlash using the SPI infrastucture
//
//  Exported functions:
//      None
//
//	Description:
//======================================================================================
// Revision history
// ================
//	02-12-2006 - created by Ronetix (www.ronetix.at)
//======================================================================================
#include <stdlib.h>
#include <string.h>
#include <cyg/io/devtab.h>                  		// definition of BLOCK_DEVTAB_TABLE and BLOCK_DEVTAB_ENTRY macro
#include <cyg/io/user_config_keys.h>
#include <pkgconf/devs_at45_dataflash_spi.h>        // configuration data
#include <cyg/io/spi.h>

HAL_SPI_EXPORTED_DEVICES


#define AT45_TIMEOUT 				100

// 0 - interrupt based SPI transfers
// 1 - polled SPI transfers
//===================================
#define	FLAG_POLLED					0


// READ COMMANDS 
//==============
#define DB_CONTINUOUS_ARRAY_READ   	0xE8		/* Continuous array read */
#define DB_BURST_ARRAY_READ        	0xE8		/* Burst array read */
#define DB_PAGE_READ               	0xD2		/* Main memory page read */
#define DB_BUF1_READ               	0xD4		/* Buffer 1 read */
#define DB_BUF2_READ               	0xD6		/* Buffer 2 read */
#define DB_STATUS                  	0xD7		/* Status Register */
#define	DB_GETID					0x9F

// PROGRAM and ERASE COMMANDS 
//===========================
#define DB_BUF1_WRITE              	0x84		/* Buffer 1 write */
#define DB_BUF2_WRITE              	0x87		/* Buffer 2 write */
#define DB_BUF1_PAGE_ERASE_PGM     	0x83		/* Buffer 1 to main memory page program with built-In erase */
#define DB_BUF1_PAGE_ERASE_FASTPGM 	0x93		/* Buffer 1 to main memory page program with built-In erase, Fast program */
#define DB_BUF2_PAGE_ERASE_PGM     	0x86		/* Buffer 2 to main memory page program with built-In erase */
#define DB_BUF2_PAGE_ERASE_FASTPGM 	0x96		/* Buffer 1 to main memory page program with built-In erase, Fast program */
#define DB_BUF1_PAGE_PGM           	0x88		/* Buffer 1 to main memory page program without built-In erase */
#define DB_BUF1_PAGE_FASTPGM       	0x98		/* Buffer 1 to main memory page program without built-In erase, Fast program*/
#define DB_BUF2_PAGE_PGM           	0x89		/* Buffer 2 to main memory page program without built-In erase */
#define DB_BUF2_PAGE_FASTPGM       	0x99		/* Buffer 1 to main memory page program without built-In erase, Fast program*/
#define DB_PAGE_ERASE              	0x81		/* Page Erase */
#define DB_BLOCK_ERASE             	0x50		/* Block Erase */
#define DB_PAGE_PGM_BUF1           	0x82		/* Main memory page through buffer 1 */
#define DB_PAGE_FASTPGM_BUF1       	0x92		/* Main memory page through buffer 1, Fast program */
#define DB_PAGE_PGM_BUF2           	0x85		/* Main memory page through buffer 2 */
#define DB_PAGE_FastPGM_BUF2       	0x95		/* Main memory page through buffer 2, Fast program */

// ADDITIONAL COMMANDS 
//====================
#define DB_PAGE_2_BUF1_TRF         	0x53		/* Main memory page to buffer 1 transfert */
#define DB_PAGE_2_BUF2_TRF         	0x55		/* Main memory page to buffer 2 transfert */
#define DB_PAGE_2_BUF1_CMP         	0x60		/* Main memory page to buffer 1 compare */
#define DB_PAGE_2_BUF2_CMP          0x61		/* Main memory page to buffer 2 compare */
#define DB_AUTO_PAGE_PGM_BUF1      	0x58		/* Auto page rewrite throught buffer 1 */
#define DB_AUTO_PAGE_PGM_BUF2      	0x59		/* Auto page rewrite throught buffer 2 */



// DataFlash channel definition
//=============================
typedef struct 
{
	cyg_mutex_t		mutex;
    
    cyg_spi_device	*spi_device;
    
	int		page_size;
	int		page_count;		
	int		chip_size;    
	
	int		nbits;
    
} AT45_CHAN;



// local definitions
//==================
static bool at45_init(struct cyg_devtab_entry *tab);
static int 	at45_lookup(struct cyg_devtab_entry **tab, struct cyg_devtab_entry *st, const char *name);
static int 	at45_bwrite(cyg_io_handle_t handle, const void *buf, cyg_uint32 *len, cyg_uint32 pos);
static int 	at45_bread(cyg_io_handle_t handle, void *buf, cyg_uint32 *len, cyg_uint32 pos);

static int 	at45_set_config(cyg_io_handle_t handle, cyg_uint32 key, const void *buffer, cyg_uint32 *len);
static int  at45_get_config(cyg_io_handle_t handle, cyg_uint32 key, void *buffer, cyg_uint32 *len);

static int 	at45_bwrite(cyg_io_handle_t handle, const void *buffer, cyg_uint32 *len, cyg_uint32 addr);
static int 	at45_bread(cyg_io_handle_t handle, void *buf, cyg_uint32 *len, cyg_uint32 pos);
static int 	at45_wait_busy( AT45_CHAN *chan );


static AT45_CHAN At45_chan;


BLOCK_DEVIO_TABLE(
                    at45_handlers,
                    at45_bwrite,
                    at45_bread,
                    NULL,
                    at45_get_config,
                    at45_set_config
                 );

BLOCK_DEVTAB_ENTRY(
                    at45_device,
                    CYGDAT_DEV_AT45_DATAFLASH_NAME,
                    NULL,
                    &at45_handlers,
                    at45_init,
                    at45_lookup,
                    &At45_chan                                  // Private data pointer
                  );




//=========================================================================================
//                      at45_init()
//                      ----------
//  Initialize EEPROM device driver
//
//=========================================================================================
static bool at45_init(struct cyg_devtab_entry *tab)
{
AT45_CHAN *chan;

    chan = (AT45_CHAN*)tab->priv;
    chan->spi_device       = cyg_spi_dataflash_dev0;

	cyg_mutex_init( &chan->mutex );

    return ENOERR;
}

//=========================================================================================
//                              at45_lookup()
//                              ------------
//  Parameter:
//      tab -> pointer to devtab entry of device
//      st  -> pointer to devtab entry of parent device
//
//  Return:
//      EX_SUCCESS - if no error
//
//  Note:
//=========================================================================================
static int at45_lookup(struct cyg_devtab_entry **tab, struct cyg_devtab_entry *st, const char *name)
{
AT45_CHAN *chan;
cyg_uint8 rx_data[8], tx_data[8];
int ii, id;

    chan = (AT45_CHAN *)(*tab)->priv;

	// read the status register to determine the chip
	//===============================================
	ii = 0;
	do
	{
		tx_data[0] = DB_STATUS;
		tx_data[1] = 0;	
		cyg_spi_transfer(chan->spi_device, FLAG_POLLED, 2, tx_data, rx_data);
		
		if ( rx_data[1] != 0xFF )
		{
			id = (rx_data[1] >> 2) & 0xF;
			break;
		}
			
	}
	while( ++ii < 10);

	if ( ii >= 10 )
		return EIO;

	switch( id )
	{
		// AT45DB011
		//==========
		case 0x3:
			chan->page_size  = 256 + 8;
			chan->page_count = 512;
			chan->nbits = 9;
			break;	
			
		// AT45DB021
		//==========
		case 0x5:	
			chan->page_size  = 256 + 8;
			chan->page_count = 1024;
			chan->nbits = 9;
			break;	
			
		// AT45DB041
		//==========
		case 0x7:	
			chan->page_size  = 256 + 8;
			chan->page_count = 2048;
			chan->nbits = 9;
			break;	
			
		// AT45DB081
		//==========
		case 0x9:	
			chan->page_size  = 256 + 8;
			chan->page_count = 4096;
			chan->nbits = 9;
			break;	
			
		// AT45DB161
		//==========
		case 0xB:	
			chan->page_size  = 512 + 16;
			chan->page_count = 4096;
			chan->nbits = 10;
			break;	
			
		// AT45DB321
		//==========
		case 0xC:	
		case 0xD:	
			chan->page_size  = 512 + 16;
			chan->page_count = 8192;
			chan->nbits = 10;
			break;	
			
		// AT45DB642
		//==========
		case 0xF:	
			chan->page_size  = 1024 + 32;
			chan->page_count = 8192;
			chan->nbits = 11;
			break;	

		// unsupported chip
		//=================			
		default:
			chan->page_size  = 0;
			chan->page_count = 0;
			chan->nbits = 0;
			break;
	}
	

	chan->chip_size = chan->page_size * chan->page_count;

#if 0
	// read the manufacturer ID
	//=========================
	memset( rx_data, 0, sizeof(rx_data) );
	memset( tx_data, 0, sizeof(tx_data) );
	
	tx_data[0] = DB_GETID;

	cyg_spi_transfer(chan->spi_device, FLAG_POLLED, 5, tx_data, rx_data);
#endif
	

    return ENOERR;
}



//=========================================================================================
//                      at45_wait_busy()
//                      ----------------
//  Parameter:
//		chan -> channel descriptor
//
//  Return:
//      ENOERR - if no error
//		EIO    - if timeout occured
//
//=========================================================================================
static int at45_wait_busy( AT45_CHAN *chan )
{
cyg_uint8 tx_cmd[2], rx_cmd[2];
cyg_tick_count_t	tm;

	tm = cyg_current_time() + AT45_TIMEOUT;

	// Issue Get Status command
	//==============================
	while(1)
	{
		tx_cmd[0] = DB_STATUS;
		tx_cmd[1] = 0;
		
		rx_cmd[0] = 0;
		rx_cmd[1] = 0;
		
		cyg_spi_transfer(chan->spi_device, FLAG_POLLED, 2, tx_cmd, rx_cmd);
		
		if ( rx_cmd[1] & 0x80 ) break;		
		if ( cyg_current_time() > tm ) break;
	}

	if ( !(rx_cmd[1] & 0x80) )
		return EIO;
	
    return ENOERR;
}



//=========================================================================================
//                      at45_bwrite()
//                      ------------
//  Parameter:
//      handle -> device handle
//      buf    -> buffer to send
//      len    -> pointer to variable with buffer's length in blocks
//      addr   -> block  number to write
//
//  On exit:
//      In *len -> number of realy sent blocks
//
//  Return:
//      ENOERR - if no error
//
//=========================================================================================
static int at45_bwrite(cyg_io_handle_t handle, const void *buffer, cyg_uint32 *len, cyg_uint32 addr)
{
int stat, ii;
cyg_uint8 cmd[4];
cyg_devtab_entry_t *t = (cyg_devtab_entry_t *)handle;
AT45_CHAN *chan;
cyg_uint32 aa;

    chan = (AT45_CHAN *)t->priv;

    // Range checking
    //===============
    if ( addr*chan->page_size + (*len)*chan->page_size > chan->chip_size )  return EINVAL;

	// Lock device
	//============
	cyg_mutex_lock( &chan->mutex );

	stat = ENOERR;
	for ( ii=0; ii < *len; ii++ )
	{
		// Wait device
		//============
		stat = at45_wait_busy( chan );
		if ( stat ) break;

		// Select device
		//==============
		cyg_spi_transaction_begin( chan->spi_device );

    	aa = ( (addr / chan->page_size) << chan->nbits) + (addr % chan->page_size);
	
		// Issue Continuous Read command
		//==============================
		cmd[0] = DB_PAGE_PGM_BUF1;
		cmd[1] = (unsigned char)((aa & 0x00FF0000) >> 16);
		cmd[2] = (unsigned char)((aa & 0x0000FF00) >> 8);
		cmd[3] = (unsigned char)(aa & 0x000000FF) ;
		
		cyg_spi_transaction_transfer( chan->spi_device, FLAG_POLLED, 4, cmd, NULL, 0);				
		cyg_spi_transaction_transfer( chan->spi_device, FLAG_POLLED, chan->page_size, buffer, NULL, 0);				
		
		(cyg_uint8 *)buffer += chan->page_size;
		addr += chan->page_size;
		
		// Deselect device
		//==============
		cyg_spi_transaction_end( chan->spi_device );
		
	}

	// Unlock device
	//==============
	cyg_mutex_unlock( &chan->mutex );

	*len = ii;
    return stat;
}


//=========================================================================================
//                          at45_bread()
//                          ------------
//  Parameter:
//      handle -> device handle
//      buffer -> pointer to data to receive
//      len    -> pointer of variable with length of data to read. On exit *len contains
//                the number of realy sended blocks
//
//  Return:
//      ENOERR - if no error
//
//=========================================================================================
static int at45_bread(cyg_io_handle_t handle, void *buffer,  cyg_uint32 *len, cyg_uint32 addr)
{
int stat, ii;
cyg_uint8 cmd[8];
cyg_devtab_entry_t *t = (cyg_devtab_entry_t *)handle;
AT45_CHAN *chan;
cyg_uint32 aa;

    chan = (AT45_CHAN *)t->priv;

    // Range checking
    //===============
    if ( addr*chan->page_size + (*len)*chan->page_size > chan->chip_size )  return EINVAL;

	// Lock device
	//============
	cyg_mutex_lock( &chan->mutex );

	// Wait device
	//============
	stat = at45_wait_busy( chan );
	if ( stat ) 	
	{
		cyg_mutex_unlock( &chan->mutex );
		return stat;
	}
	

	for(ii=0; ii < *len; ii++ )
	{
		// Select device
		//==============
		cyg_spi_transaction_begin( chan->spi_device );		
		
    	aa = ( (addr / chan->page_size) << chan->nbits) + (addr % chan->page_size);
	
		// Issue Continuous Read command
		//==============================
		cmd[0] = DB_PAGE_READ;
		cmd[1] = (unsigned char)((aa & 0x00FF0000) >> 16);
		cmd[2] = (unsigned char)((aa & 0x0000FF00) >> 8);
		cmd[3] = (unsigned char)(aa & 0x000000FF) ;
		cmd[4] = 0;
		cmd[5] = 0;
		cmd[6] = 0;
		cmd[7] = 0;

		cyg_spi_transaction_transfer( chan->spi_device, FLAG_POLLED, 8, cmd, NULL, 0);				
		cyg_spi_transaction_transfer( chan->spi_device, FLAG_POLLED, chan->page_size, NULL, buffer, 0);				

		(cyg_uint8 *)buffer += chan->page_size;
		addr += chan->page_size;
		
		// Deselect device
		//================
		cyg_spi_transaction_end( chan->spi_device );		
		
	}

	// Unlock device
	//==============
	cyg_mutex_unlock( &chan->mutex );

    return stat;
}



//===================================================================================================
//                              at45_set_config()
//                              ----------------
//  Parameter:
//      handle -> channel handle
//      key    -> CYG_IO_CONFIG_SPI_DIVIDER
//      buffer -> pointer to configuration data
//      len    -> pointer to variable that contain the size of buffer. MUST be exactly the size of
//                parameter's structure
//
//  Rerturn:
//      ENOERR - if no error
//      EINVAL - if the size of buffer don't mach the size of parameter's structure
//
//  Description:
//      Set diverse config parameters
//===================================================================================================
static int at45_set_config(cyg_io_handle_t handle, cyg_uint32 key, const void *buffer, cyg_uint32 *len)
{
cyg_devtab_entry_t *t = (cyg_devtab_entry_t *)handle;
AT45_CHAN *chan;

    chan = (AT45_CHAN *)t->priv;

	switch ( key )
	{
	
		default:
			return ENOSYS;
			break;		
	}		

	
    return ENOERR;
}

//===================================================================================================
//                              at45_get_config()
//                              ----------------
//  Parameter:
//      handle -> channel handle
//      key    -> CYG_IO_CONFIG_SPI_CLOCK
//      buffer -> pointer to configuration data
//      len    -> pointer to variable that contain the size of buffer. MUST be exactly the size of
//                parameter's structure
//
//  Rerturn:
//      ENOERR - if no error
//      EINVAL - if the size of buffer don't mach the size of parameter's structure
//
//  Description:
//      Return diverse config parameters
//===================================================================================================
static int at45_get_config(cyg_io_handle_t handle, cyg_uint32 key, void *buffer, cyg_uint32 *len)
{
cyg_devtab_entry_t *t = (cyg_devtab_entry_t *)handle;
AT45_CHAN *chan;

    chan = (AT45_CHAN *)t->priv;

	switch ( key )
	{
		case CYG_IO_CONFIG_SIZE:
			if (*len != sizeof(int))
				return EINVAL;
				
			*((int*)buffer) = chan->page_size * chan->page_count;
			break;
			
		case CYG_IO_CONFIG_BLOCK_SIZE:
			if (*len != sizeof(int))
				return EINVAL;
				
			*((int*)buffer) = chan->page_size;
			break;			
		
		default:
			return ENOSYS;
			break;		
	}		

    return ENOERR;
}

