//==========================================================================
//                  nand.cxx
//                  --------
//
//  Generic NAND Flash driver
//
//	www.ronetix.at
//==========================================================================
#include <pkgconf/system.h>
#include <cyg/io/devtab.h>
#include <cyg/io/user_config_keys.h>
#include <pkgconf/devs_flash_nand.h>
#include "nandio.h"


static bool nand_init(struct cyg_devtab_entry *tab);
static Cyg_ErrNo nand_bwrite(cyg_io_handle_t handle, const void *buf, cyg_uint32 *len, cyg_uint32 pos);
static Cyg_ErrNo nand_bread(cyg_io_handle_t handle, void *buf, cyg_uint32 *len, cyg_uint32 pos);
static Cyg_ErrNo nand_lookup(struct cyg_devtab_entry **tab, struct cyg_devtab_entry *sub_tab, const char *name);
static Cyg_ErrNo nand_get_config(cyg_io_handle_t handle,  cyg_uint32 key, void *buf, cyg_uint32 *len);
static Cyg_ErrNo nand_set_config(cyg_io_handle_t handle, cyg_uint32 key, const void *buf, cyg_uint32* len);


BLOCK_DEVIO_TABLE( nand_handlers,
                   nand_bwrite,
                   nand_bread,
                   NULL,                    // select is not used
                   nand_get_config,
                   nand_set_config
                 );


#ifdef CYGPKG_DEVS_FLASH_NAND_0
    static class NAND sm0 CYG_INIT_PRIORITY( DRIVERS );
	BLOCK_DEVTAB_ENTRY(nand_io0,
                  CYGDAT_DEVS_FLASH_NAND_NAME_0,
                  CYGDAT_DEVS_FLASH_NAND_LOW_LEVEL_NAME_0, 			// Base device name
                  &nand_handlers,
                  nand_init,
                  nand_lookup,
                  &sm0);                                 			// Private data pointer
#endif

#ifdef CYGPKG_DEVS_FLASH_NAND_1
    static class NAND sm1 CYG_INIT_PRIORITY( DRIVERS );
	BLOCK_DEVTAB_ENTRY(nand_io1,
                  CYGDAT_DEVS_FLASH_NAND_NAME_1,
                  CYGDAT_DEVS_FLASH_NAND_LOW_LEVEL_NAME_1,           // Base device name
                  &nand_handlers,
                  nand_init,
                  nand_lookup,
                  &sm1);                                 			// Private data pointer
#endif



//=============================================================================
//                      	nand_lookup()
//                      	-------------
//=============================================================================
static bool nand_init(struct cyg_devtab_entry *tab)
{
    return ENOERR;
}

//=============================================================================
//                      	nand_lookup()
//                      	-------------
//  Parameter:
//      dev     -> device tab entry
//      sub_tab -> sub device tab entry
//      name    -> name of device
//
//  Return:
//      ENOERR     - if no error
//
//  Description:
//      Look for Nand Flash device and build a bad blocks table.
//=============================================================================
static Cyg_ErrNo nand_lookup(struct cyg_devtab_entry **tab, struct cyg_devtab_entry *sub_tab,const char *name)
{
class NAND *chan = (class NAND *)(*tab)->priv;
int stat;

	chan->nandio_handle = (cyg_io_handle_t)sub_tab;
	
    stat = chan->init();
    if (stat==ENODEV) return stat;

    return ENOERR;
}

//=============================================================================
//                          nand_bwrite()
//                          -------------
//  Parameter:
//      handle - channel descriptor
//		buff_data  -> buffer for the page data
//		buff_spare -> buffer for the spare data
//		page_num   -> page number to read
//
//  Return:
//      ENOERR     - if no error
//
//  Description:
//      Write a page to the NAND Flash device
//=============================================================================
static Cyg_ErrNo nand_bwrite(cyg_io_handle_t handle, const void *buffers, cyg_uint32 *len, cyg_uint32 page_num)
{
cyg_devtab_entry_t *tab = (cyg_devtab_entry_t *)handle;
class NAND  *chan = (class NAND  *)tab->priv;
cyg_uint32 *bb;
cyg_uint8 *buff_data, *buff_spare;

	bb = (cyg_uint32 *)buffers;

	buff_data  = (cyg_uint8*)bb[0];
	buff_spare = (cyg_uint8*)bb[1];

    return chan->page_write(page_num, (cyg_uint8 *)buff_data, (cyg_uint8 *)buff_spare );
}

//=============================================================================
//                          nand_bread()
//                          ------------
//  Parameter:
//      handle - channel descriptor
//		buff_data  -> buffer for the page data
//		buff_spare -> buffer for the spare data
//		page_num   -> page number to read
//
//  Return:
//      ENOERR - if no error
//
//  Description:
//      Read a page
//=============================================================================
static Cyg_ErrNo nand_bread(cyg_io_handle_t handle, void *buffers, cyg_uint32 *len, cyg_uint32 page_num)
{
cyg_devtab_entry_t *tab = (cyg_devtab_entry_t *)handle;
class NAND *chan = (class NAND  *)tab->priv;
cyg_uint8 *buff_data, *buff_spare;

	buff_data  = (cyg_uint8*)(*((cyg_uint32*)buffers));
	buff_spare = (cyg_uint8*)(*((cyg_uint32*)buffers+1));

    return chan->page_read(page_num, (cyg_uint8 *)buff_data, (cyg_uint8 *)buff_spare );
}


//=============================================================================
//                              nand_get_config()
//                              -----------------
//  Parameter:
//      handle - device descriptor
//      key    - key selector
//      buff   - buffer for config data
//      len    - lenght of buffer.
//
//  Return:
//      ENOERR  - if no errors
//      EINVAL - if *len dosn't match config structure
//
//  Description:
//      Get NAND Flash configuration data
//=============================================================================
static Cyg_ErrNo nand_get_config(cyg_io_handle_t handle,  cyg_uint32 key, void *buf, cyg_uint32 *len)
{
cyg_devtab_entry_t *tab = (cyg_devtab_entry_t *)handle;
class NAND  *chan = (class NAND  *)tab->priv;

	if ( *len != 4 )
		return EINVAL;

    switch ( key )
    {
        // Get Nand Flash parameter
        //=========================
        case CYG_IO_CONFIG_FLASH_NAND_NUM_BLOCKS:	
        	*(cyg_uint32 *)buf = chan->info->num_blocks;
            break;

        case CYG_IO_CONFIG_FLASH_NAND_PAGES_PER_BLOCK:	
        	*(cyg_uint32 *)buf = chan->info->pages_per_block;
            break;

        case CYG_IO_CONFIG_FLASH_NAND_PAGE_SIZE:	
        	*(cyg_uint32 *)buf = chan->info->page_size;
            break;


        // unknow key selector
        //====================
        default : 
        	return EINVAL;
    }

    return ENOERR;
}



//=============================================================================
//						nand_set_config()
//						-----------------
//
//  Parameter:
//      handle - device descriptor
//      key    - key selector
//      buff   - buffer for config data
//      len    - lenght of buffer.
//
//  Return:
//      ENOERR  - if no errors
//      EINVAL - if *len dosn't match config structure
//
//  Description:
//      Set NAND Flash configuration data
//=============================================================================
static Cyg_ErrNo nand_set_config(cyg_io_handle_t handle, cyg_uint32 key, const void *buf, cyg_uint32 *len)
{
cyg_devtab_entry_t *tab = (cyg_devtab_entry_t *)handle;
class NAND *chan = (class NAND  *)tab->priv;
int stat, nn;

	if ( *len != sizeof(int) )
		return EINVAL;

	nn = *(int*)buf;

	switch ( key )
	{
		// Erase one block
		//================
		case CYG_IO_CONFIG_FLASH_NAND_ERASE_BLOCK:
			stat = chan->block_erase( nn );	
			if ( stat ) return stat;
			break;

		case CYG_IO_CONFIG_FLASH_NAND_MARK_BAD_BLOCK:
			stat = chan->mark_bad_block( nn );
			if ( stat ) return stat;			
			break;
					
		default:	
			return EINVAL;	
	}
			
    return ENOERR;
}




