/*
 * YAFFS: Yet another FFS. A NAND-flash specific file system.
 * yaffs_ecos.c  The interface functions for using YAFFS via the eCos interface.
 *
 * Created by Ilko Iliev
 * www.ronetix.at
 *
 */
#include <pkgconf/system.h>
#include <pkgconf/hal.h>
#include <pkgconf/kernel.h>
#include <pkgconf/io_fileio.h>
#include <unistd.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <errno.h>
#include <dirent.h>
#include <cyg/fileio/fileio.h>
#include <cyg/kernel/kapi.h>
#include <stdlib.h>
#include <string.h>
#include <cyg/io/io.h>
#include <cyg/io/user_config_keys.h>
 
 
#include "yaffs_guts.h"
#include "yaffsfs.h"
#include "yaffscfg.h"
#include <string.h>
#include "yportenv.h"
#include "yaffs_packedtags1.h"
#include "yaffs_packedtags2.h"

#define	YAFFS_ECOS_MAX_PATH				YAFFS_MAX_NAME_LENGTH

// Interface bwtween YAFFS and eCos fileio system
//================================================
static int yaffs_ecos_mount(  cyg_fstab_entry *fste, cyg_mtab_entry *mte );
static int yaffs_ecos_umount( cyg_mtab_entry *mte );
static int yaffs_ecos_open(   cyg_mtab_entry *mte, cyg_dir dir, const char *name, int mode,  cyg_file *file );
static int yaffs_ecos_unlink( cyg_mtab_entry *mte, cyg_dir dir, const char *name );
static int yaffs_ecos_mkdir(  cyg_mtab_entry *mte, cyg_dir dir, const char *name );
static int yaffs_ecos_rmdir(  cyg_mtab_entry *mte, cyg_dir dir, const char *name);
static int yaffs_ecos_rename( cyg_mtab_entry *mte, cyg_dir dir1, const char *name1, cyg_dir dir2, const char *name2 );
static int yaffs_ecos_link(   cyg_mtab_entry *mte, cyg_dir dir1, const char *name1, cyg_dir dir2, const char *name2, int type );
static int yaffs_ecos_opendir( cyg_mtab_entry *mte, cyg_dir dir, const char *name, cyg_file *file );
static int yaffs_ecos_chdir(  cyg_mtab_entry *mte, cyg_dir dir, const char *name, cyg_dir *dir_out );
static int yaffs_ecos_stat(   cyg_mtab_entry *mte, cyg_dir dir, const char *name, struct stat *buf);
static int yaffs_ecos_getinfo( cyg_mtab_entry *mte, cyg_dir dir, const char *name, int key, void *buf, int len );
static int yaffs_ecos_setinfo( cyg_mtab_entry *mte, cyg_dir dir, const char *name, int key, void *buf, int len );

static int yaffs_ecos_fo_dirread(struct CYG_FILE_TAG *fp, struct CYG_UIO_TAG *uio);
static int yaffs_ecos_fo_dirlseek(struct CYG_FILE_TAG *fp, off_t *pos, int whence);
static int yaffs_ecos_fo_closedir(struct CYG_FILE_TAG *fp);


static int yaffs_ecos_fo_read( struct CYG_FILE_TAG *fp,  struct CYG_UIO_TAG *uio);
static int yaffs_ecos_fo_write(struct CYG_FILE_TAG *fp, struct CYG_UIO_TAG *uio);
static int yaffs_ecos_fo_lseek(struct CYG_FILE_TAG *fp, off_t *apos, int whence );
static int yaffs_ecos_fo_ioctl(struct CYG_FILE_TAG *fp, CYG_ADDRWORD com, CYG_ADDRWORD data);
static int yaffs_ecos_fo_fsync(struct CYG_FILE_TAG *fp, int mode );
static int yaffs_ecos_fo_fstat(struct CYG_FILE_TAG *fp, struct stat *buf );
static int yaffs_ecos_fo_close(struct CYG_FILE_TAG *fp);
static int yaffs_ecos_fo_getinfo(struct CYG_FILE_TAG *fp, int key, void *buf, int len );
static int yaffs_ecos_fo_setinfo(struct CYG_FILE_TAG *fp, int key, void *buf, int len );


// Interface between YAFFS and the NAND Flash block driver
//=========================================================
static int yaffs_ecos_nand_init(yaffs_Device *dev);
static int yaffs_ecos_nand_write( yaffs_Device *dev,int chunkInNAND, const __u8 *data, const yaffs_ExtendedTags *tags);
static int yaffs_ecos_nand_read( yaffs_Device *dev,int chunkInNAND, __u8 *data, yaffs_ExtendedTags *tags);
static int yaffs_ecos_nand_block_erase( yaffs_Device *dev, int block_num );
static int yaffs_ecos_nand_block_markbad( yaffs_Device *dev, int block_num );
static int yaffs_ecos_nand_block_query(yaffs_Device *dev, int block_num, yaffs_BlockState *state, int *sequenceNumber);



//==========================================================================
// Filesystem table entries

// -------------------------------------------------------------------------
// Fstab entry.

FSTAB_ENTRY(yaffs_fste, "yaffs", 0,
            CYG_SYNCMODE_FILE_FILESYSTEM|CYG_SYNCMODE_IO_FILESYSTEM,
            yaffs_ecos_mount,
            yaffs_ecos_umount,
            yaffs_ecos_open,
            yaffs_ecos_unlink,
            yaffs_ecos_mkdir,
            yaffs_ecos_rmdir,
            yaffs_ecos_rename,
            yaffs_ecos_link,
            yaffs_ecos_opendir,
            yaffs_ecos_chdir,
            yaffs_ecos_stat,
            yaffs_ecos_getinfo,
            yaffs_ecos_setinfo);


// -------------------------------------------------------------------------
// File operations.

static cyg_fileops yaffs_fileops =
{
    yaffs_ecos_fo_read,
    yaffs_ecos_fo_write,
    yaffs_ecos_fo_lseek,
    yaffs_ecos_fo_ioctl,
    cyg_fileio_seltrue,
    yaffs_ecos_fo_fsync,
    yaffs_ecos_fo_close,
    yaffs_ecos_fo_fstat,
    yaffs_ecos_fo_getinfo,
    yaffs_ecos_fo_setinfo
};


// -------------------------------------------------------------------------
// Directory file operations.

static cyg_fileops yaffs_dirops =
{
    yaffs_ecos_fo_dirread,
    (cyg_fileop_write *)  cyg_fileio_enosys,
    yaffs_ecos_fo_dirlseek,
    (cyg_fileop_ioctl *)  cyg_fileio_enosys,
    cyg_fileio_seltrue,
    (cyg_fileop_fsync *)  cyg_fileio_enosys,
    yaffs_ecos_fo_closedir,
    (cyg_fileop_fstat *)  cyg_fileio_enosys,
    (cyg_fileop_getinfo *)cyg_fileio_enosys,
    (cyg_fileop_setinfo *)cyg_fileio_enosys
};



typedef struct
{
	int flag_busy;
    cyg_io_handle_t     		dev_h;			// device handle for the block device
    
    char						*yaffs_dev_prefix;
    yaffs_Device				yaffs_dev;
	int fd;	
	char 						curr_dir[YAFFS_ECOS_MAX_PATH];
	
} YAFFS_ECOS_CHAN;


static YAFFS_ECOS_CHAN YaffseCosChan[YAFFSFS_N_PARTITIONS];
static int YaffseCosChan_idx=0;	

unsigned yaffs_traceMask = YAFFS_TRACE_SCAN |  YAFFS_TRACE_GC /*| YAFFS_TRACE_GC_DETAIL | YAFFS_TRACE_WRITE */ | YAFFS_TRACE_ERASE | YAFFS_TRACE_TRACING | YAFFS_TRACE_ALLOCATE | YAFFS_TRACE_CHECKPOINT;
//unsigned yaffs_traceMask = 0xFFFFFFFF; 


//=============================================================================
//						chan_alloc()
//						------------
// 
//=============================================================================
static YAFFS_ECOS_CHAN *chan_alloc(void)
{
static int flag_initialized=false;	
YAFFS_ECOS_CHAN *chan;

	if ( flag_initialized == false )
	{
		memset( YaffseCosChan, 0, sizeof(YaffseCosChan) ); 
		flag_initialized = true;
	}
	
	if( YaffseCosChan_idx >= YAFFSFS_N_PARTITIONS )
		return NULL;
	
	chan = &YaffseCosChan[YaffseCosChan_idx++];
	chan->flag_busy = true;
	
	return chan;
}

//=============================================================================
//						chan_free()
//						-----------
// 
//=============================================================================
static int chan_free(YAFFS_ECOS_CHAN *chan )
{
	if ( YaffseCosChan_idx <= 0 )
		return EINVAL;
		
	chan->flag_busy = false;
	YaffseCosChan_idx--;	
	
	return ENOERR;
}


void yaffsfs_SetError(int err)
{
	//Do whatever to set error
	errno = err;
}

void yaffsfs_Lock(void)
{
}

void yaffsfs_Unlock(void)
{
}

__u32 yaffsfs_CurrentTime(void)
{
	return (__u32)cyg_current_time();
}

void yaffsfs_LocalInitialisation(void)
{
	// Define locking semaphore.
}


//=============================================================================
//							get_param()
//							-----------
//	Parameter:
//		line  -> the input string
//		param -> parameter name
//		ret_value -> pointer to variable which receives the value
//
//	Return:
//		ENOERR - if no errors
//		EINVAL - invalid format
//		ENOENT - no such parameter
//
//	Description:
//		Get a parameter from a string:
//		line = " start=1, end=128";
//		param = "start";
//		*ret_value = 1;
//=============================================================================
static int get_param( char *line, char *param, unsigned int *ret_value )
{
char *cc, *dd;

	*ret_value = 0;

	cc = strstr( line, param );
	if ( cc == NULL ) return ENOENT;
		
	// look for '='
	//=============		
	while( *cc && *cc != '=' )
		cc++;
		
	if ( *cc != '=' ) return EINVAL;
		
	cc++;	
		
	*ret_value = strtol( cc, &dd, 0 );		
		
	if ( dd == cc ) return EINVAL;
		
	return ENOERR;				
}

//=============================================================================
//						yaffs_ecos_mount()
//						------------------
// Process a mount request.
//=============================================================================
static int yaffs_ecos_mount( cyg_fstab_entry *fste, cyg_mtab_entry *mte )
{
YAFFS_ECOS_CHAN *chan;
char devname[32], *dd = devname, *pp;
int stat, len, num_blocks, block_start=0, block_end=0;
int flag_format=false;

	memset(devname, 0, sizeof(devname) );
	pp = (char *)mte->devname;
	
	// skip all spaces
	//=================
	while( *pp && *pp != ' ' )		
		*dd++ = *pp++;	
		
	*dd = 0;	
	if ( *pp == ' ' )
	{
		// we have additional parameters
		//==============================
		stat = get_param( pp, "start", &block_start );
		if ( stat && stat != ENOENT ) return EINVAL;	
		
		stat = get_param( pp, "end", &block_end );
		if ( stat && stat != ENOENT ) return EINVAL;	
		
		stat = get_param( pp, "format", &flag_format );
		if ( stat && stat != ENOENT ) return EINVAL;	
	}

	chan = chan_alloc();
	if ( chan == NULL ) return EIO;

	stat = ENOERR;	
	while(1)
	{	
	    stat = cyg_io_lookup(devname, &chan->dev_h);
	    if (stat) break;
	
		len = 4;
		stat = cyg_io_get_config( chan->dev_h, CYG_IO_CONFIG_FLASH_NAND_PAGE_SIZE, &chan->yaffs_dev.nBytesPerChunk, &len );
		if ( stat ) break;
		
		len = 4;
		stat = cyg_io_get_config( chan->dev_h, CYG_IO_CONFIG_FLASH_NAND_NUM_BLOCKS, &num_blocks, &len );
		if ( stat ) break;
		
		len = 4;
		stat = cyg_io_get_config( chan->dev_h, CYG_IO_CONFIG_FLASH_NAND_PAGES_PER_BLOCK, &chan->yaffs_dev.nChunksPerBlock, &len );
		if ( stat ) break;

		// range checking
		//===============
		if ( block_start > num_blocks )
		{
			stat = EINVAL;
			break;
		}
			
		// range checking
		//===============
		if ( block_end == 0 || block_end > num_blocks )
			block_end = num_blocks;

		// Format if necessary
		//====================
		if ( flag_format )
		{
			int ii;
			
			for ( ii=block_start; ii < block_end; ii++ )
			{
				len = 4;
				cyg_io_set_config( chan->dev_h, CYG_IO_CONFIG_FLASH_NAND_ERASE_BLOCK, &ii, &len );
			}
		}
	
		chan->yaffs_dev.isYaffs2 = (chan->yaffs_dev.nBytesPerChunk > 526) ? true : false ;
		chan->yaffs_dev.nReservedBlocks = 5;
		chan->yaffs_dev.nCheckpointReservedBlocks = 5;
		chan->yaffs_dev.startBlock 		= block_start; 
		chan->yaffs_dev.endBlock 		= block_end; 
		chan->yaffs_dev.wideTnodesDisabled=0;
		chan->yaffs_dev.nShortOpCaches = 10; 			// Use caches
		chan->yaffs_dev.genericDevice = (void *)0;	// Used to identify the device in fstat.
		chan->yaffs_dev.priv = chan;
		chan->yaffs_dev_prefix = mte->name;
		
		chan->yaffs_dev.initialiseNAND			= yaffs_ecos_nand_init;
		chan->yaffs_dev.writeChunkWithTagsToNAND  = yaffs_ecos_nand_write;
		chan->yaffs_dev.readChunkWithTagsFromNAND = yaffs_ecos_nand_read;
		chan->yaffs_dev.eraseBlockInNAND		= yaffs_ecos_nand_block_erase;
		chan->yaffs_dev.markNANDBlockBad		= yaffs_ecos_nand_block_markbad;
		chan->yaffs_dev.queryNANDBlock			= yaffs_ecos_nand_block_query;
		
		stat = yaffs_partition_add( chan->yaffs_dev_prefix, &chan->yaffs_dev );
		if ( stat ) break;

		stat = yaffs_mount( chan->yaffs_dev_prefix );
		if ( stat ) break;
		
		stat = ENOERR;
		break;
	}

	if ( stat )
	{
		chan_free( chan );
		return stat;
	}
	
	mte->root = (cyg_dir)mte->name;
	strncpy( chan->curr_dir, mte->name, sizeof(chan->curr_dir)-1);
		
	mte->data = (CYG_ADDRWORD)chan; 
	
	return ENOERR;	
}


//=============================================================================
//						yaffs_ecos_umount()
//						-------------------
// Unmount the filesystem. This will currently only succeed if the
// filesystem is empty.
//=============================================================================
static int yaffs_ecos_umount( cyg_mtab_entry *mte )
{
YAFFS_ECOS_CHAN *chan;
int stat;

	chan = (YAFFS_ECOS_CHAN *)mte->data;
	
	stat = yaffs_unmount( chan->yaffs_dev_prefix );
	if ( stat ) return stat;
	
	stat = yaffs_partition_remove( chan->yaffs_dev_prefix );
	if ( stat ) return stat;
	
	chan_free( chan );
        
    return ENOERR;
}


//=============================================================================
//						yaffs_ecos_open()
//						-----------------
// Open a file for reading or writing.
//=============================================================================
static int yaffs_ecos_open( cyg_mtab_entry *mte, cyg_dir dir, const char *name, int mode,  cyg_file *file )
{
YAFFS_ECOS_CHAN *chan;
int fd;
char fname[YAFFS_ECOS_MAX_PATH];

	chan = (YAFFS_ECOS_CHAN *)mte->data;

	snprintf( fname, sizeof(fname)-1, "%s/%s", chan->curr_dir, name );

	fd = yaffs_open( fname, mode, S_IREAD | S_IWRITE );
	if ( fd < 0 ) return EIO;
	
    // Initialize the file object
    
    file->f_flag        |= mode & CYG_FILE_MODE_MASK;
    file->f_type        = CYG_FILE_TYPE_FILE;
    file->f_ops         = &yaffs_fileops;

	file->f_offset = 0;
    file->f_data        = (CYG_ADDRWORD)fd;
    file->f_xops        = 0;
	
	return ENOERR;
}



//=============================================================================
//						yaffs_ecos_ulink()
//						-----------------
// Remove a file link from its directory.
//=============================================================================
static int yaffs_ecos_unlink( cyg_mtab_entry *mte, cyg_dir dir, const char *name )
{
	
	return ENOERR;	
}


//=============================================================================
//						yaffs_ecos_mkdir()
//						------------------
// Create a new directory.
//=============================================================================
static int yaffs_ecos_mkdir( cyg_mtab_entry *mte, cyg_dir dir, const char *name )
{
int ret;	
char dirname[YAFFS_ECOS_MAX_PATH];

	snprintf( dirname, sizeof(dirname), "%s/%s", (char *)dir, name );
	
	ret = yaffs_mkdir( dirname, 0);
	if ( ret ) return ret;
	
	return ENOERR;
}



//=============================================================================
//						yaffs_ecos_rmdir()
//						------------------
// Remove a directory.
//=============================================================================
static int yaffs_ecos_rmdir(cyg_mtab_entry *mte, cyg_dir dir, const char *name)
{
int ret;	
char dirname[YAFFS_ECOS_MAX_PATH];

	snprintf( dirname, sizeof(dirname), "%s/%s", (char *)dir, name );

	ret = yaffs_rmdir( dirname );
	if ( ret ) return ret;

	return ENOERR;
}	



//=============================================================================
//						yaffs_ecos_rename()
//						-------------------
// Rename a file/dir.
//=============================================================================
static int yaffs_ecos_rename( cyg_mtab_entry *mte, cyg_dir dir1, const char *name1, cyg_dir dir2, const char *name2 )
{
int ret;	
char dirname1[YAFFS_ECOS_MAX_PATH];
char dirname2[YAFFS_ECOS_MAX_PATH];

	snprintf( dirname1, sizeof(dirname1), "%s/%s", (char *)dir1, name1 );
	snprintf( dirname2, sizeof(dirname2), "%s/%s", (char *)dir2, name2 );

	ret = yaffs_rename( dirname1, dirname2);
	if ( ret ) return ret;
	
	return ENOERR;
}


//=============================================================================
//						yaffs_ecos_link()
//						-----------------
// Make a new directory entry for a file.
//=============================================================================
static int yaffs_ecos_link( cyg_mtab_entry *mte, cyg_dir dir1, const char *name1, cyg_dir dir2, const char *name2, int type )
{
int ret;	
char dirname1[YAFFS_ECOS_MAX_PATH];
char dirname2[YAFFS_ECOS_MAX_PATH];

	snprintf( dirname1, sizeof(dirname1), "%s/%s", (char *)dir1, name1 );
	snprintf( dirname2, sizeof(dirname2), "%s/%s", (char *)dir2, name2 );
	
	ret = yaffs_link( dirname1, dirname2 );
	if ( ret ) return ret;
	
	return ENOERR;
}


//=============================================================================
//						yaffs_ecos_opendir()
//						--------------------
// Open a directory for reading.
//=============================================================================
static int yaffs_ecos_opendir( cyg_mtab_entry *mte, cyg_dir dir, const char *name, cyg_file *file )
{
YAFFS_ECOS_CHAN *chan;
yaffs_DIR *yaffs_dir;
int stat;
char dirname[YAFFS_ECOS_MAX_PATH];
char *curr_dir = (char *)dir;

	chan = (YAFFS_ECOS_CHAN *)mte->data;

	snprintf( dirname, sizeof(dirname), "%s/%s", curr_dir, name );

	yaffs_dir = yaffs_opendir( dirname );

    // Initialize the file object
	//===========================
    file->f_type    = CYG_FILE_TYPE_FILE;
    file->f_ops     = &yaffs_dirops;
    file->f_data    = (CYG_ADDRWORD)yaffs_dir;
    file->f_xops    = 0;
    file->f_offset  = 0;


	return ENOERR;
}


//=============================================================================
//						yaffs_ecos_chdir()
//						------------------
// Change directory support - not implemented yet
//=============================================================================
static int yaffs_ecos_chdir( cyg_mtab_entry *mte, cyg_dir dir, const char *name, cyg_dir *dir_out )
{
YAFFS_ECOS_CHAN *chan = (YAFFS_ECOS_CHAN *)mte->data;

	return ENOERR;
}


//=============================================================================
//						yaffs_ecos_stat()
//						-----------------
// Get struct stat info for named object.
//=============================================================================
static int yaffs_ecos_stat( cyg_mtab_entry *mte, cyg_dir dir, const char *name, struct stat *buf)
{
YAFFS_ECOS_CHAN *chan = (YAFFS_ECOS_CHAN *)mte->data;
int ret;
struct yaffs_stat y_stat;
char dirname[YAFFS_ECOS_MAX_PATH];

	snprintf( dirname, sizeof(dirname)-1, "%s/%s", chan->curr_dir, name );

	ret = yaffs_stat( dirname, &y_stat);
	if ( ret ) 
	{
		yaffsfs_SetError(EACCES);
		return ENOENT;
	}

    // Fill in the status
    //===================
    buf->st_mode        = y_stat.st_mode;
    buf->st_ino         = y_stat.st_ino;
    buf->st_dev         = y_stat.st_dev;
    buf->st_nlink       = y_stat.st_dev;
    buf->st_uid         = y_stat.st_uid;
    buf->st_gid         = y_stat.st_gid;
    buf->st_size        = y_stat.st_size;
    buf->st_atime       = y_stat.yst_atime;
    buf->st_mtime       = y_stat.yst_mtime;
    buf->st_ctime       = y_stat.yst_ctime;

	return ENOERR;
}

//=============================================================================
//						yaffs_ecos_getinfo()
//						--------------------
// Get struct stat info for named object.
//=============================================================================
static int yaffs_ecos_getinfo( cyg_mtab_entry *mte, cyg_dir dir, const char *name, int key, void *buf, int len )
{
	return EINVAL;
}


//=============================================================================
//						yaffs_ecos_setinfo()
//						--------------------
// Setinfo. Nothing to support here at present.
//=============================================================================
static int yaffs_ecos_setinfo( cyg_mtab_entry *mte, cyg_dir dir, const char *name, int key, void *buf, int len )
{
    // No setinfo keys supported at present
    return EINVAL;
}




//=============================================================================
//						yaffs_ecos_fo_dirread()
//						-----------------------
// Read a single directory entry from a file.
//=============================================================================
static int yaffs_ecos_fo_dirread(struct CYG_FILE_TAG *fp, struct CYG_UIO_TAG *uio)
{
YAFFS_ECOS_CHAN *chan = (YAFFS_ECOS_CHAN *) fp->f_mte->data;
yaffs_DIR *yaffs_dir = (yaffs_DIR *) fp->f_data;
off_t pos = fp->f_offset;
int err = ENOERR;
struct dirent *ent = (struct dirent *)uio->uio_iov[0].iov_base;
yaffs_dirent *y_dirent;

	y_dirent = yaffs_readdir(yaffs_dir);
	if ( y_dirent == NULL ) return EEOF;

	strncpy( ent->d_name, y_dirent->d_name, sizeof(ent->d_name)-1 );
	
    uio->uio_resid         -= sizeof(struct dirent);
    fp->f_offset++;

	return ENOERR;
}





//=============================================================================
//						yaffs_ecos_fo_dirlseek()
//						------------------------
// Seek directory to start.
//=============================================================================
static int yaffs_ecos_fo_dirlseek(struct CYG_FILE_TAG *fp, off_t *pos, int whence)
{
yaffs_DIR *yaffs_dir = (yaffs_DIR *) fp->f_data;

    if (whence != SEEK_SET || *pos != 0)
        return EINVAL;
   
	yaffs_rewinddir(yaffs_dir);
	
	fp->f_offset = 0;
	
    return ENOERR;
}



//=============================================================================
//						yaffs_ecos_fo_closedir()
//						---------------------
// Close a file or a directory
//=============================================================================
static int yaffs_ecos_fo_closedir(struct CYG_FILE_TAG *fp)
{    
yaffs_DIR *yaffs_dir = (yaffs_DIR *) fp->f_data;
    
	yaffs_closedir( yaffs_dir );
	    
    return ENOERR;
}






//=============================================================================
//						yaffs_eocs_fo_read()
//						--------------------
// Read data from the file.
//=============================================================================
static int yaffs_ecos_fo_read(struct CYG_FILE_TAG *fp, struct CYG_UIO_TAG *uio)
{
int ret, ii;
cyg_uint32 ll;
ssize_t resid = uio->uio_resid;
	
	
    // Loop over the io vectors until there are none left

    for (ii = 0; ii < uio->uio_iovcnt; ii++)
    {
        cyg_iovec  *iov  = &uio->uio_iov[ii];
        char       *buf  = (char *) iov->iov_base;
        off_t       len  = iov->iov_len;

        // Loop over each vector filling it with data from the file
        
        while (len > 0)
        {
			ll = len;
			
			ret = yaffs_read( fp->f_data, buf, ll );
            if (ret < 0)
                return EIO;

			if ( ret == 0 )
				break;

            len   -= ll;
            buf   += ll;
            resid -= ll;
        }
    }

    // We successfully read some data, update the access time, 
    // file offset and transfer residue

    uio->uio_resid     = resid;
	
	return ENOERR;
}



//=============================================================================
//						yaffs_ecos_fo_write()
//						---------------------
// Write data to file.
//=============================================================================
static int yaffs_ecos_fo_write(struct CYG_FILE_TAG *fp, struct CYG_UIO_TAG *uio)
{
int ret, ii;
ssize_t resid = uio->uio_resid;
	

    // Now loop over the iovecs until they are all done, or we get an error
	//=====================================================================    
    for (ii = 0; ii < uio->uio_iovcnt; ii++)
    {
        cyg_iovec  *iov  = &uio->uio_iov[ii];
        char       *buf  = (char *) iov->iov_base;
        off_t       len  = iov->iov_len;
 
        // Loop over the vector writing it to the file 
        // until it has all been done
		//============================================
        while (len > 0)
        {
            cyg_uint32 ll = len;

			ret = yaffs_write(fp->f_data, buf, ll);

            len   -= ll;
            buf   += ll;
            resid -= ll;

            // Stop writing if there is no more space in the file
            if (ret != ll )
            {
            	uio->uio_resid = resid;
            	return ENOSPC;
            }
            
            if (ret < 0)
                return EIO;
        }
    }        

    uio->uio_resid = resid;
	
	return ENOERR;	
}


//=============================================================================
//						yaffs_ecos_fo_lseek()
//						---------------------
// Seek to a new file position.
//=============================================================================
static int yaffs_ecos_fo_lseek(struct CYG_FILE_TAG *fp, off_t *apos, int whence )
{
off_t pos;
	
	pos = yaffs_lseek( fp->f_data, *apos, whence);
	if ( pos < 0 ) return EIO;
	
	*apos = pos;
	
	return ENOERR;
}	


//=============================================================================
//						yaffs_ecos_fo_ioctl()
//						---------------------
// Handle ioctls. Currently none are defined.
//=============================================================================
static int yaffs_ecos_fo_ioctl(struct CYG_FILE_TAG *fp, CYG_ADDRWORD com, CYG_ADDRWORD data)
{
    // No Ioctls currenly defined.

    return EINVAL;
}


//=============================================================================
//						yaffs_ecos_fo_fsync()
//						---------------------
// Force the file out to data storage.
//=============================================================================
static int yaffs_ecos_fo_fsync(struct CYG_FILE_TAG *fp, int mode )
{
    // Data is always permanently where it belongs, nothing to do
    // here.
  
	return ENOERR; 
}


//=============================================================================
//						yaffs_ecos_fo_close()
//						---------------------
// Close a file. We just decrement the refcnt and let it go away if
// that is all that is keeping it here.
//=============================================================================
static int yaffs_ecos_fo_close(struct CYG_FILE_TAG *fp)
{
int err;	

	err = yaffs_close(fp->f_data);
    return (err < 0 ) ? EIO : ENOERR;
}


//=============================================================================
//						yaffs_ecos_fo_fstat()
//						---------------------
// Get file status.
//=============================================================================
static int yaffs_ecos_fo_fstat(struct CYG_FILE_TAG *fp, struct stat *buf )
{
int ret;
struct yaffs_stat y_stat;

	ret = yaffs_fstat( fp->f_data, &y_stat );
	if ( ret < 0 ) return EIO;

    // Fill in the status
    //===================
    buf->st_mode        = y_stat.st_mode;
    buf->st_ino         = y_stat.st_ino;
    buf->st_dev         = y_stat.st_dev;
    buf->st_nlink       = y_stat.st_dev;
    buf->st_uid         = y_stat.st_uid;
    buf->st_gid         = y_stat.st_gid;
    buf->st_size        = y_stat.st_size;
    buf->st_atime       = y_stat.yst_atime;
    buf->st_mtime       = y_stat.yst_mtime;
    buf->st_ctime       = y_stat.yst_ctime;

	return ENOERR;
}


//=============================================================================
//						yaffs_ecos_fo_getinfo()
//						-----------------------
// Get info. Currently only supports fpathconf().
//=============================================================================
static int yaffs_ecos_fo_getinfo(struct CYG_FILE_TAG *fp, int key, void *buf, int len )
{
	return EINVAL;
}




//=============================================================================
//						yaffs_ecos_fo_getinfo()
//						-----------------------
// Set info. Nothing supported here.
//=============================================================================
static int yaffs_ecos_fo_setinfo(struct CYG_FILE_TAG *fp, int key, void *buf, int len )
{
    // No setinfo key supported at present
    
    return EINVAL;
}




//=========================================================================================================
//	Functions to communicate with the NAND block driver	
//=========================================================================================================
typedef union
{
	yaffs_PackedTags1 tag_page_small;
	yaffs_PackedTags2 tag_page_large;
	char buff_spate[64];
	
} YAFFS_PACKET_TAGS;

//=============================================================================
//						yaffs_ecos_tags_pack()
//						-----------------------
// extended_tag -> packed_tag
//=============================================================================
void yaffs_ecos_tags_pack( int flag_large_page, void *packet_tags, const void *ex_tags )
{
	if ( flag_large_page == true )
		yaffs_PackTags2( (yaffs_PackedTags2 *)packet_tags, (yaffs_ExtendedTags *)ex_tags );
	else
		yaffs_PackTags1( (yaffs_PackedTags1 *)packet_tags, (yaffs_ExtendedTags *)ex_tags );
}

//=============================================================================
//						yaffs_ecos_tags_unpack()
//						------------------------
//	packed_tag -> extended_tag
//=============================================================================
void yaffs_ecos_tags_unpack( int flag_large_page, void *packet_tags, void *ex_tags )
{
	if ( flag_large_page == true )
		yaffs_UnpackTags2( (yaffs_ExtendedTags *)ex_tags, (yaffs_PackedTags2 *)packet_tags );
	else
		yaffs_UnpackTags1( (yaffs_ExtendedTags *)ex_tags, (yaffs_PackedTags1 *)packet_tags );
}


//=============================================================================
//						yaffs_ecos_nand_init()
//						------------------------
//
//=============================================================================
static int yaffs_ecos_nand_init(yaffs_Device *dev)
{
	return YAFFS_OK;
}


//=============================================================================
//						yaffs_ecos_nand_write()
//						------------------------
//
//=============================================================================
static int yaffs_ecos_nand_write( yaffs_Device *dev,int chunkInNAND,const __u8 *data, const yaffs_ExtendedTags *tags)
{
YAFFS_ECOS_CHAN *chan = (YAFFS_ECOS_CHAN *)dev->priv;	
YAFFS_PACKET_TAGS pack_tag;
cyg_uint8 *buff[2];
cyg_uint32 len=1;
int stat;

	buff[0] = (cyg_uint8 *)data;
	
	if (tags)
	{
		yaffs_ecos_tags_pack( dev->isYaffs2, &pack_tag, tags );
		buff[1] = (cyg_uint8*)&pack_tag;
	}
	else
		buff[1] = NULL;
	
	stat = cyg_io_bwrite( chan->dev_h, buff, &len, chunkInNAND );
	return (stat) ? YAFFS_FAIL : YAFFS_OK;
}





//=============================================================================
//						yaffs_ecos_nand_read()
//						-----------------------
//
//
//=============================================================================
static int yaffs_ecos_nand_read( yaffs_Device *dev, int chunkInNAND, __u8 *data, yaffs_ExtendedTags *tags)
{
YAFFS_ECOS_CHAN *chan = (YAFFS_ECOS_CHAN *)dev->priv;	
YAFFS_PACKET_TAGS pack_tag;
cyg_uint8 *buff[2];
cyg_uint32 len=1;
int stat;

	buff[0] = (cyg_uint8 *)data;
	
	if (tags)
		buff[1] = (cyg_uint8 *)&pack_tag;
	else
		buff[1] = NULL;	
	
	stat = cyg_io_bread( chan->dev_h, buff, &len, chunkInNAND );
	if ( stat ) return YAFFS_FAIL;

	
	if (tags)
		yaffs_ecos_tags_unpack( dev->isYaffs2, &pack_tag, tags );

	return ENOERR;
}

//=============================================================================
//						yaffs_ecos_nand_block_erase()
//						-----------------------------
//
//
//=============================================================================
static int yaffs_ecos_nand_block_erase( yaffs_Device *dev, int block_num )
{
YAFFS_ECOS_CHAN *chan = (YAFFS_ECOS_CHAN *)dev->priv;	
cyg_uint32 len = 4;
int stat;

	stat = cyg_io_set_config( chan->dev_h, CYG_IO_CONFIG_FLASH_NAND_ERASE_BLOCK, &block_num, &len );	
	return (stat) ? YAFFS_FAIL : YAFFS_OK;
}


//=============================================================================
//						yaffs_ecos_nand_block_markbad()
//						-------------------------------
//
//
//=============================================================================
static int yaffs_ecos_nand_block_markbad( yaffs_Device *dev, int block_num )
{
YAFFS_ECOS_CHAN *chan = (YAFFS_ECOS_CHAN *)dev->priv;	
cyg_uint32 len = 4;
int stat;
	
	diag_printf("** warning: yaffs_ecos_nand_block_markbad( block_num = %d )\r\n", block_num );
	
	stat = cyg_io_set_config( chan->dev_h, CYG_IO_CONFIG_FLASH_NAND_MARK_BAD_BLOCK, &block_num, &len );
	return (stat) ? YAFFS_FAIL : YAFFS_OK;
}


//=============================================================================
//						yaffs_ecos_nand_block_query()
//						------------------------------
//
//
//=============================================================================
static int yaffs_ecos_nand_block_query(yaffs_Device *dev, int block_num, yaffs_BlockState *state, int *sequenceNumber)
{
yaffs_ExtendedTags tags;

	*sequenceNumber = 0;
	
	yaffs_ecos_nand_read(dev, block_num, NULL, &tags);
	if (tags.blockBad)
	{
		*state = YAFFS_BLOCK_STATE_DEAD;
	}
	else 
	if (!tags.chunkUsed)
	{
		*state = YAFFS_BLOCK_STATE_EMPTY;
	}
	else 
	if (tags.chunkUsed)
	{
		*state = YAFFS_BLOCK_STATE_NEEDS_SCANNING;
		*sequenceNumber = tags.sequenceNumber;
	}
	
	return YAFFS_OK;
}
