//=================================================================
//
//        usb_msd.c
//
// USB Mass Storage device
//
//	www.ronetix.at
//==========================================================================


#include <stdio.h>
#include <cyg/infra/cyg_ass.h>
#include <cyg/infra/diag.h>
#include <cyg/kernel/kapi.h>
#include <cyg/hal/hal_arch.h>
#include <cyg/io/io.h>
#include <cyg/io/usb/usbs.h>
#include <cyg/io/usb/usb_msd.h>
#include <cyg/io/user_config_keys.h>
#include "msd.h"

#ifndef CYGPKG_DEVS_DISK_AT91_MCI
	#define MSD_DUMMY
#endif	

typedef enum 
{
    sdcValid=0,                 // Everything is golden
    sdcCardInitCommFailure,     // Communication has never been established with card
    sdcCardNotInitFailure,      // Card did not go into an initialization phase
    sdcCardInitTimeout,         // Card initialization has timedout
    sdcCardTypeInvalid,         // Card type was not able to be defined
    sdcCardBadCmd,              // Card did not reconized the command
    sdcCardTimeout,             // Card timedout during a read, write or erase sequence   
    sdcCardCRCError,            // A CRC error has occurred during a read, data should be invalidated
    sdcCardDataRejected,         // Card and data sent's CRC did not match
    sdcEraseTimedOut            // Erase took longer than it should have 
    
} SDC_Error;



#define	DUMP_LEVEL	0


#if (DUMP_LEVEL > 0)
	#define DUMP_MSG(level, _format_, _args_...)  { if (DUMP_LEVEL >= level) diag_printf(_format_, ## _args_);} 
#else
	#define DUMP_MSG
#endif		


#define	SENSE_SET_SENSE_KEY( base_array, value )			{base_array[2] = value & 0xF;}
#define	SENSE_SET_ASC( base_array, value )					{base_array[12] = value;}
#define	SENSE_SET_ASCQ( base_array, value )					{base_array[13] = value;}


typedef struct
{
	int id;
	char *name;
		
} ID_NAME;

static const ID_NAME UfiCommands[] = 
{
	{ SCSI_INQUIRY, 				"INQUIRY" },
	{ SCSI_READ_FORMAT_CAPACITY,	"READ_FORMAT_CAPACITY" },
	{ SCSI_READ_CAPACITY,			"READ_CAPACITY" },
	{ SCSI_READ_10,					"READ_10" }, 
	{ SCSI_WRITE_10,				"WRITE_10" },
	{ SCSI_REQUEST_SENSE,			"REQUEST_SENSE" }, 
	{ SCSI_MODE_SELECT,				"MODE_SELECT" }, 
	{ SCSI_MODE_SENSE,				"MODE_SENSE" }, 
	{ SCSI_PREVENT_ALLOW_MEDIUM_REMOVAL, "PREVENT_ALLOW_MEDIUM_REMOVAL" },
	{ SCSI_TEST_UNIT_READY, 		"TEST_UNIT_READY" },		
	{ SCSI_VERIFY,					"VERIFY" },
	{ SCSI_STOP_START,				"STOP_START" },

	// must be the last
	//=================	
	{ -1, "UNKNOW" }
};



static const ID_NAME MsdStates[] = 
{
	{ USB_MSD_STATE_WAIT, 			"STATE_WAIT" },
	{ USB_MSD_STATE_DATA_IN, 		"STATE_DATA_IN" },
	{ USB_MSD_STATE_DATA_OUT, 		"STATE_DATA_OUT" },
	{ USB_MSD_STATE_DATA_IN_END, 	"STATE_DATA_IN_END" },

	// must be the last
	//=================	
	{ -1, "UNKNOW" }
};

static MSD_CHAN Msd_chans[USB_MSD_MAX_DEVICES];


static void msd_reset_sense_data( MSD_CHAN *chan );
static int usb_msd_send( MSD_CHAN *chan, void *buff, int len );
static int usb_msd_send_status( MSD_CHAN *chan, int status, cyg_uint32 residue );

static bool SDC_DetectCard(void);
static int SDC_SectorRead( MSD_CHAN *chan, cyg_uint32 sector_addr, cyg_uint8 *buffer );
static int SDC_SectorWrite( MSD_CHAN *chan, cyg_uint32 sector_addr, cyg_uint8* buffer );

static void usb_msd_task( cyg_addrword_t data );


// Standard Response to INQUIRY command
// ====================================
const InquiryResponse inq_resp = 
{
	0x00,		// peripheral device is connected, direct access block device
	0x80,       // removable
	0x04,	 	// version = 00=> does not conform to any standard, 4=> SPC-2
	0x02,		// response is in format specified by SPC-2
	0x20,		// n-4 = 36-4=32= 0x20
	0x00,		// sccs etc.
	0x00,		// bque=1 and cmdque=0,indicates simple queueing 00 is obsolete,
			    // but as in case of other device, we are just using 00
	0x00,		// 00 obsolete, 0x80 for basic task queueing
	"Microchp",	// this is the T10 assigned Vendor ID
	"Mass Storage    ",
	"0001"
};





// ********************************************************************************



// ----------------------------------------------------------------------------
// The enumeration data.
//
// For simplicity this configuration involves just a single interface.
// The target has to list all the endpoints, or the Linux kernel will
// not allow application code to access them. Hence the information
// provided by the device drivers has to be turned into endpoint descriptors.

static usb_configuration_descriptor usb_configuration = 
{
    length:             USB_CONFIGURATION_DESCRIPTOR_LENGTH,					// 0x9
    type:               USB_CONFIGURATION_DESCRIPTOR_TYPE,						// 0x2
    total_length_lo:    USB_CONFIGURATION_DESCRIPTOR_TOTAL_LENGTH_LO(1, 2),		// 9+9 + 2x7 = 0x20
    total_length_hi:    USB_CONFIGURATION_DESCRIPTOR_TOTAL_LENGTH_HI(1, 2),		// 
    number_interfaces:  1,
    configuration_id:   1,                                                      // id 0 is special according to the spec
    configuration_str:  0,                                                      // no string defined
    attributes:         USB_CONFIGURATION_DESCRIPTOR_ATTR_REQUIRED | USB_CONFIGURATION_DESCRIPTOR_ATTR_SELF_POWERED,
    max_power:          50                                                      // 
};

static usb_interface_descriptor usb_interface = 
{
    length:             USB_INTERFACE_DESCRIPTOR_LENGTH,
    type:               USB_INTERFACE_DESCRIPTOR_TYPE,
    interface_id:       0,
    alternate_setting:  0,
    number_endpoints:   2,
    interface_class:    USB_DEVICE_DESCRIPTOR_CLASS_MASS_STORAGE,
    interface_subclass: USB_DEVICE_DESCRIPTOR_SUBCLASS_SCSI,
    interface_protocol: USB_INTERFACE_DESCRIPTOR_PROTOCOL_BULK_ONLY_TRANSPORT,
    interface_str:      0
};

static const char *usb_strings[] = 
{
    "\x04\003\x09\004",
    
    "\x10\003R\000e\000d\000 \000H\000a\000t\000",
    "\x1C\003R\000e\000d\000 \000H\000a\000t\000 \000e\000C\000o\000s\000 \000",
    "\x14\0031\0002\0003\0004\0005\0006\0007\0008\0009\000",
    "\x22\0030\0001\0002\0003\0004\0005\0006\0007\0008\0009\000A\000B\000C\000D\000E\000F\000"
};


static const usb_endpoint_descriptor usb_endpoints[] = 
{
    {
        length:         USB_ENDPOINT_DESCRIPTOR_LENGTH,
        type:           USB_ENDPOINT_DESCRIPTOR_TYPE,
        endpoint:       USB_ENDPOINT_DESCRIPTOR_ENDPOINT_OUT | 1,		// direction HOST -> DEVICE
        attributes:     USB_ENDPOINT_DESCRIPTOR_ATTR_BULK,
        max_packet_lo:  64,
        max_packet_hi:  0,
        interval:       0
    },
    {
        length:         USB_ENDPOINT_DESCRIPTOR_LENGTH,
        type:           USB_ENDPOINT_DESCRIPTOR_TYPE,
        endpoint:       USB_ENDPOINT_DESCRIPTOR_ENDPOINT_IN | 2,		// direction DEVICE -> HOST
        attributes:     USB_ENDPOINT_DESCRIPTOR_ATTR_BULK,
        max_packet_lo:  64,
        max_packet_hi:  0,
        interval:       0
    },
    {
        length:         USB_ENDPOINT_DESCRIPTOR_LENGTH,
        type:           USB_ENDPOINT_DESCRIPTOR_TYPE,
        endpoint:       USB_ENDPOINT_DESCRIPTOR_ENDPOINT_IN | 3,		// direction DEVICE -> HOST
        attributes:     USB_ENDPOINT_DESCRIPTOR_ATTR_INTERRUPT,
        max_packet_lo:  8,
        max_packet_hi:  0,
        interval:       10
    }
};


static usbs_enumeration_data usb_enum_data = 
{
    {
        length:                 USB_DEVICE_DESCRIPTOR_LENGTH,
        type:                   USB_DEVICE_DESCRIPTOR_TYPE,
        usb_spec_lo:            USB_DEVICE_DESCRIPTOR_USB11_LO,
        usb_spec_hi:            USB_DEVICE_DESCRIPTOR_USB11_HI,
        device_class:           0x00,//USB_DEVICE_DESCRIPTOR_CLASS_MASS_STORAGE,
        device_subclass:        0x00,//USB_DEVICE_DESCRIPTOR_SUBCLASS_SCSI,
        device_protocol:        0x00,//USB_INTERFACE_DESCRIPTOR_PROTOCOL_BULK_ONLY_TRANSPORT,
        max_packet_size:        8,
        vendor_lo:              0x52,   // Note: this is not an allocated vendor id
        vendor_hi:              0x42,
        product_lo:             0x00,
        product_hi:             0x01,
        device_lo:              0x00,
        device_hi:              0x01,
        manufacturer_str:       1,
        product_str:            2,
        serial_number_str:      3,
        number_configurations:  1
    },
    total_number_interfaces:    1,
    total_number_endpoints:     3,
    total_number_strings:       4,
    configurations:             &usb_configuration,
    interfaces:                 &usb_interface,
    endpoints:                  usb_endpoints,
    strings:                    usb_strings 
};






//=============================================================================
//						dump_buff()
//						-----------
//=============================================================================
static void dump_buff( cyg_uint8 *buff, int len )
{
int ii, nn, offset;	
	
	nn = 0;
	offset = 0;
	
	for( ii=0; ii < len; ii++)
	{
		if ( nn == 0 )
			diag_printf("%03X: ", offset );
			
		diag_printf("%02X ", *buff++ );
		if ( nn == 7 )
			diag_printf("- ");
			
		if ( ++nn == 16 )
		{
			diag_printf("\r\n");	
			nn = 0;
		}
		
		offset++;
	}
	
	diag_printf("\r\n");
}

static void dump_buff_e( cyg_uint8 *buff, int len, int ndisp )
{
int ii, nn, offset;	
	
	nn = 0;
	offset = 0;
	
	for( ii=0; ii < len; ii++)
	{
		if ( nn == 0 )
			diag_printf("%03X: ", offset );
			
		diag_printf("%02X ", *buff++ );
		if ( nn+1 == (ndisp/2) )
			diag_printf("- ");
			
		if ( ++nn == ndisp )
		{
			diag_printf("\r\n");	
			nn = 0;
		}
		
		offset++;
	}
	
	diag_printf("\r\n");
}

//=============================================================================
//						get_ufi_name()
//						--------------
//=============================================================================
static char *get_ufi_name( int id )
{
const ID_NAME *pp = UfiCommands;

	while( pp->id != -1 )
	{
		if ( pp->id == id )
			break;
			
		pp++;	
	}
	
	return pp->name;
}

//=============================================================================
//						get_state_name()
//						----------------
//=============================================================================
static char *get_state_name( int id )
{
const ID_NAME *pp = MsdStates;

	while( pp->id != -1 )
	{
		if ( pp->id == id )
			break;
			
		pp++;	
	}
	
	return pp->name;
}


//=============================================================================
//				    MSD_ClassControlHandler()
//					-------------------------
//	Parameter:
//		            endpoint ->
//		            callback_data ->
//
//	Return:
//		
//	Description:
//		            This function deals with the MASS Storage device controls
//=============================================================================
static usbs_control_return MSD_ClassControlHandler(usbs_control_endpoint* endpoint, void* callback_data)
{
//MSD_CHAN *chan = (MSD_CHAN *)callback_data;	
usbs_control_return result = USBS_CONTROL_RETURN_STALL;
usb_devreq* devreq = (usb_devreq*) endpoint->control_buffer;
int direction;

    direction   = devreq->type & USB_DEVREQ_DIRECTION_MASK;

    if(direction & 0x80)
    {
        DUMP_MSG(6, "<- ");
    }
    else
    {
        DUMP_MSG(6, "-> ");
    }

    if (RESET_REQUEST == devreq->request) 
    {
    	DUMP_MSG(5, "RESET_REQUEST\r\n");
    	endpoint->buffer_size = 0;
        result = USBS_CONTROL_RETURN_HANDLED;
    } 
    else 
    if (MAX_LUN_REQUEST == devreq->request) 
    {
    	DUMP_MSG(5, "MAX_LUN_REQUEST\r\n");
		endpoint->buffer[0] = MAX_LUN;
    	endpoint->buffer_size 	= MAX_LUN_LENGTH;
		result = USBS_CONTROL_RETURN_HANDLED;
    } 
    else
    {
       	DUMP_MSG( 0, "unknow class control request, devreq->request = %d\r\n", devreq->request);
    }

    return result;
}

//=============================================================================
//							MSD_EndpointStateChange()
//							-------------------------
//
//=============================================================================
static void MSD_EndpointStateChange(usbs_control_endpoint* endpoint, void* callback_data, usbs_state_change new_state, int old_state)
{
MSD_CHAN *chan = (MSD_CHAN *)callback_data;
     
    if ( new_state & 0x80 )
    {
        new_state = USBS_STATE_CHANGE_SUSPENDED;
        endpoint->state = USBS_STATE_POWERED;
    }     
    
    switch(new_state)
    {
        case USBS_STATE_CHANGE_DETACHED:
            DUMP_MSG(5, " DETACHED ");
            break;
            
        case USBS_STATE_CHANGE_ATTACHED:
            DUMP_MSG(5, " ATTACHED ");
            break;
            
        case USBS_STATE_CHANGE_POWERED:
            DUMP_MSG(5, " POWERED ");
            break;
            
        case USBS_STATE_CHANGE_RESET:
            DUMP_MSG(5, " RESET ");  
            break;
            
        case USBS_STATE_CHANGE_ADDRESSED:
            DUMP_MSG(5, " ADDRESSED ");
            break;
            
        case USBS_STATE_CHANGE_CONFIGURED:
            DUMP_MSG(5, " CONFIGURED ");
            break;
            
        case USBS_STATE_CHANGE_DECONFIGURED:
            DUMP_MSG(5, " DECONFIGURED ");    
            break;
            
        case USBS_STATE_CHANGE_SUSPENDED:
            DUMP_MSG(5, " SUSPENDED ");
            chan->msd_state = USB_MSD_STATE_WAIT;
//            msd_init();				
            return;
            
        case USBS_STATE_CHANGE_RESUMED:
            DUMP_MSG(5, " RESUMED ");       
            break;
    }

    // if the state changes to configured, start the OUT endpoint (ready to receive)
    // =============================================================================
    if ( new_state == USBS_STATE_CHANGE_CONFIGURED && old_state != USBS_STATE_CHANGE_CONFIGURED )
    {
        usbs_start_rx(chan->conf.endp_out);
    }
}


//=============================================================================
//							MSD_BulkOutCallback()
//							---------------------
//	Parameter:
//		chan->
//
//	Return:
//		ENOERR
//
//	Description:
//		Deal with the packets HOST -> DEVICE
//=============================================================================
static void MSD_BulkOutCallback( void* arg, int result )
{
MSD_CHAN *chan = (MSD_CHAN *)arg;	
    
    DUMP_MSG(7, "MSD_BulkOutCallback( result = %d, LUN = %d )\r\n", result, chan->msd_cbw.bCBWLUN );

	if ( result < 0 ) 
	{
		DUMP_MSG(0, "MSD_BulkOutCallback( result = %d, LUN = %d )\r\n", result, chan->msd_cbw.bCBWLUN );
		return;
	}

    cyg_semaphore_post(&chan->completion_wait);
}

//=============================================================================
//							MSD_BulkInCallback()
//							--------------------
//	Parameter:
//		chan->
//
//	Return:
//		ENOERR
//
//	Description:
//		Deal with the packets DEVICE --> HOST
//=============================================================================
static void MSD_BulkInCallback( void *arg, int result )
{
MSD_CHAN *chan = (MSD_CHAN *)arg;	
    
    DUMP_MSG(7, "MSD_BulkInCallback( result = %d )\r\n", result );
    
    if (result == -EIO)                  // (-5) I/O Error 
    {
    	DUMP_MSG(0, "MSD_BulkInCallback( result = %d )\r\n", result );
    	
        chan->conf.endp_in->buffer            = chan->msd_buffer;
        chan->conf.endp_in->buffer_size       = 10;
        usbs_start_tx(chan->conf.endp_in);  
    }
    else 
    if (result == -EPIPE)            // broken pipe (304)
    {
    	DUMP_MSG(0, "MSD_BulkInCallback( result = %d )\r\n", result );
    }
    else 
    if (result == -EAGAIN)          // (-11) endpoint has been halted
    {
    	DUMP_MSG(0, "MSD_BulkInCallback( result = %d )\r\n", result );
        chan->conf.endp_in->buffer            = chan->msd_buffer;
        chan->conf.endp_in->buffer_size       = 10;
        usbs_start_tx(chan->conf.endp_in);        
    }

	if ( result < 0 )
	{
		DUMP_MSG(0, "MSD_BulkInCallback( result = %d )\r\n", result );
		return; 
	}

    cyg_semaphore_post(&chan->completion_wait);
}



static bool MSD_IsValidCBW( MSD_CHAN *chan )
{
	return ( (chan->msd_cbw_copy.dCBWSignature[0] != (cyg_uint8)(CBW_SIGNATURE >>  0)) ||
		     (chan->msd_cbw_copy.dCBWSignature[1] != (cyg_uint8)(CBW_SIGNATURE >>  8)) ||
		     (chan->msd_cbw_copy.dCBWSignature[2] != (cyg_uint8)(CBW_SIGNATURE >> 16))  ||
		     (chan->msd_cbw_copy.dCBWSignature[3] != (cyg_uint8)(CBW_SIGNATURE >> 24)) ) ? false : true;
}

static bool MSD_IsMeaningfulCBW( MSD_CHAN *chan )
{
	/*  3msb bits of CBWCBLength are reserved and must be 0,
     *  4msb bits of CBWLUN	are reserved and must be 0
	 *  valid CBWCBLength is between 1 and 16B
	 *  In bCBWFlags only msb indicates data direction rest must be 0
	 */  
	if( (chan->msd_cbw_copy.bCBWLUN <= 0x0f) && 
	    (chan->msd_cbw_copy.bCBWCBLength <= 0x10) &&
	    (chan->msd_cbw_copy.bCBWCBLength >= 0x01) &&
	    (chan->msd_cbw_copy.bCBWFlags==0x00 || chan->msd_cbw_copy.bCBWFlags==0x80)
	    ) 
		return true;
	else 
	    return false;	
}


//=============================================================================
// Function:        void msd_prepare_CSW()
//
// Overview:        This prepares the Status data of CSW by copying the
//                  dCSWTag from CBWTage and sets the signature 
//                  of valid CSW=53425355h
//
//=============================================================================
static void msd_prepare_CSW( MSD_CHAN *chan, int status, cyg_uint32 residue )
{ 
	DUMP_MSG(7, "msd_prepare_CSW()\r\n");

	// build the CSW structure 
	//========================
	chan->msd_csw.dCSWSignature[0] = (cyg_uint8)(CSW_SIGNATURE);     		// 0x55; 
	chan->msd_csw.dCSWSignature[1] = (cyg_uint8)(CSW_SIGNATURE >> 8);     	// 0x53; 
	chan->msd_csw.dCSWSignature[2] = (cyg_uint8)(CSW_SIGNATURE >> 16);     	// 0x42; 
	chan->msd_csw.dCSWSignature[3] = (cyg_uint8)(CSW_SIGNATURE >> 24);     	// 0x53; 

	chan->msd_csw.dCSWTag[0] = chan->msd_cbw_copy.dCBWTag[0];
	chan->msd_csw.dCSWTag[1] = chan->msd_cbw_copy.dCBWTag[1];
	chan->msd_csw.dCSWTag[2] = chan->msd_cbw_copy.dCBWTag[2];
	chan->msd_csw.dCSWTag[3] = chan->msd_cbw_copy.dCBWTag[3];

	chan->msd_csw.dCSWDataResidue[0] = (cyg_uint8)(residue);
	chan->msd_csw.dCSWDataResidue[1] = (cyg_uint8)(residue >> 8);
	chan->msd_csw.dCSWDataResidue[2] = (cyg_uint8)(residue >> 16);
	chan->msd_csw.dCSWDataResidue[3] = (cyg_uint8)(residue >> 24);

	chan->msd_csw.bCSWStatus = status;
}


 

//=============================================================================
// 						cdb_inquiry()
//						-------------
//
// Overview:        This function prepares the response of the Inquiry command
//                  A fixed Inquiry response is copied from ROM to the
//		            msd_buffer and CSWStatus, CSWDataResidue values are set
//
// Note:            None
//=============================================================================
static void cdb_inquiry( MSD_CHAN *chan )
{
USBMS_INQUIRY *inquiry = (USBMS_INQUIRY *)chan->msd_cbw_copy.CBWCB;
int len;

	len = sizeof(InquiryResponse);
	
	DUMP_MSG(5, "cdb_inquiry(op_code = 0x%X, inquire_len = %d)\r\n", inquiry->operation_code, len );

	// send data
	//==========
	usb_msd_send( chan, (cyg_uint8*)&inq_resp, len );
}

//=============================================================================
//						cdb_read_capacity()
//						-------------------
//	Parameter:
//		chan ->
//
//	Return:
//
//	Description:
//		Processes the rbc READ_CAPACITY command
//=============================================================================
static void cdb_read_capacity( MSD_CHAN *chan )
{
cyg_uint32 last_page;

	last_page = chan->mmc_num_pages -1;

	chan->msd_buffer[0] = (cyg_uint8)(last_page >> 24);
	chan->msd_buffer[1] = (cyg_uint8)(last_page >> 16);
	chan->msd_buffer[2] = (cyg_uint8)(last_page >> 8);
	chan->msd_buffer[3] = (cyg_uint8)last_page;
	
	chan->msd_buffer[4] = (cyg_uint8)(chan->mmc_page_size >> 24);
	chan->msd_buffer[5] = (cyg_uint8)(chan->mmc_page_size >> 16);
	chan->msd_buffer[6] = (cyg_uint8)(chan->mmc_page_size >> 8);
	chan->msd_buffer[7] = (cyg_uint8)chan->mmc_page_size;

	usb_msd_send( chan, chan->msd_buffer, 8 );
}

//=============================================================================
//						cdb_read_format_capacity()
//						--------------------------
//	Parameter:
//		chan -> 
//
//	Return:
//		None
//
//	Description:
//		Prepare the data response for "READ FORMAT CAPACITIES Command: 23h"
//
// usbmass-ufi10.pdf, page 33
//=============================================================================
static void cdb_read_format_capacity( MSD_CHAN *chan )
{
cyg_uint32 nn;
	
	DUMP_MSG(5, "CDB_ReadFormatCapacity()\r\n");

	// capacity list header
	//=====================
	chan->msd_buffer[0] = 0;						// reserved
	chan->msd_buffer[1] = 0;						// reserved
	chan->msd_buffer[2] = 0;						// reserved
	chan->msd_buffer[3] = 0x08;						// capacity list length	
	
	nn = chan->mmc_num_pages;
	
	// capacity descriptor
	//====================
	chan->msd_buffer[4]  = (cyg_uint8)(nn >> 24);
	chan->msd_buffer[5]  = (cyg_uint8)(nn >> 16);
	chan->msd_buffer[6]  = (cyg_uint8)(nn >> 8);
	chan->msd_buffer[7]  = (cyg_uint8)nn;
	chan->msd_buffer[8]  = 0x03;									// bit[1..0] - descriptor type
	chan->msd_buffer[9]  = (cyg_uint8)(chan->mmc_page_size >> 16);	// page length
	chan->msd_buffer[10] = (cyg_uint8)(chan->mmc_page_size >> 8);	// page length
	chan->msd_buffer[11] = (cyg_uint8)chan->mmc_page_size;			// page length
	
	usb_msd_send( chan, chan->msd_buffer, 12 );
}

//=============================================================================
// 					cdb_request_sense()
//					-------------------
//
// Overview:        This function prepares the Sense Data in response
//                  to the Request Sense Command The contents of structure
//                  REQ_SENSE_RESP are copied to msd_buffer and a
//                  success bCSWStatus=0x00 is set.
//=============================================================================
static void cdb_request_sense( MSD_CHAN *chan )
{
int ii, nn;

	DUMP_MSG(5, "cdb_request_sense()\r\n");

	nn = sizeof(chan->sense_data);
	if ( nn > chan->msd_cbw_copy.CBWCB[4] )
		nn = chan->msd_cbw_copy.CBWCB[4];
	
	for (ii=0; ii < nn; ii++)
		chan->msd_buffer[ii] = chan->sense_data[ii];

	
	usb_msd_send( chan, chan->msd_buffer, nn );
}

//=============================================================================
// Function:        void cdb_test_unit_ready()
//
// Overview:        This function prepares response to Test Unit Ready Command
//                  No data response is expected, only a CSW is to be sent 
//                  Based on the current state of the SDCard an error or
//                  success status value is set
//
//=============================================================================
static void cdb_test_unit_ready( MSD_CHAN *chan )
{
	DUMP_MSG(5, "cdb_test_unit_ready()\r\n");

	msd_reset_sense_data(chan);
	
	usb_msd_send_status( chan, GOOD_STATUS, 0 );
	
}


//=============================================================================
//						cdb_mode_select()
//						-----------------
//=============================================================================
static void cdb_mode_select( MSD_CHAN *chan )
{
	DUMP_MSG(5, "cdb_mode_select()\r\n");
		
}


//=============================================================================
// 						cdb_mode_sense()
//						----------------
//
// Overview:        This function prepares response to the Mode Sense command
//                  a basic response is implemented in this version of the code
//                  00h imples no other mode pages and 0x03 is the size of the 
//                  data (in bytes) that follows.
//
//=============================================================================
static int cdb_mode_sense( MSD_CHAN *chan )
{
USBMS_MODE_SENSE *mode_sense = (USBMS_MODE_SENSE *)chan->msd_cbw_copy.CBWCB;	
MODE_SENSE_PARAMETER sense;	

	DUMP_MSG(5, "cdb_mode_sense( aaloc_len = %d)\r\n", mode_sense->allocation_length );
	
 	// Fill the structure
 	//===================
  	if (mode_sense->allocation_length < sizeof(MODE_SENSE_PARAMETER))
  	{
	    sense.head.mode_data_length = (cyg_uint8)(mode_sense->allocation_length-sizeof(sense.head.mode_data_length));
  	}
  	else
  	{
	    sense.head.mode_data_length = (cyg_uint8)(sizeof(MODE_SENSE_PARAMETER)-sizeof(sense.head.mode_data_length));
  	}

  	sense.head.medium_type = 0;
  	sense.head.device_specific_parameter = 0;
  	sense.head.block_descriptor_length = 0;

  	if( (mode_sense->page_code == 0x06)   		// Reserved 
    	||(mode_sense->page_code == 0x3F) ) 	// Return all pages
  	{
	    sense.page.pageCode       = 0x06;
	    sense.page.reserved1      = 0x01;
	    sense.page.pageLength     = 0x0b;
	    sense.page.writeCacheDisable = 0x01; 	// entire command is treated 
	
	    sense.page.logicalBlockSize[0] = (cyg_uint8)(chan->mmc_page_size);
	    sense.page.logicalBlockSize[1] = (cyg_uint8)(chan->mmc_page_size >> 8);
	
	    sense.page.numberOfLogicalBlocks[0] = (cyg_uint8)chan->mmc_num_pages;
	    sense.page.numberOfLogicalBlocks[1] = (cyg_uint8)(chan->mmc_num_pages >> 8);
	    sense.page.numberOfLogicalBlocks[2] = (cyg_uint8)(chan->mmc_num_pages >> 16);
	    sense.page.numberOfLogicalBlocks[3] = (cyg_uint8)(chan->mmc_num_pages >> 24);
	    sense.page.numberOfLogicalBlocks[4] = 0x00;
	    
	    sense.page.powerPerformance  = 0x00; 	// 0x00 energie, 0xFF perf 
	    sense.page.lockd     = 0x00;
	    sense.page.formatd   = 0x00;
	    sense.page.writed    = 0x00;
	    sense.page.readd   	 = 0x00;
	    sense.page.reserved4 = 0x00;
	
	    // Send the Mode Sense data, always the same zeroed struct 
	    //=========================================================
	    int nn = mode_sense->allocation_length;
	    if ( nn > sizeof(MODE_SENSE_PARAMETER) )
	    	nn = sizeof(MODE_SENSE_PARAMETER);
	    
	    usb_msd_send( chan, (char *)&sense, nn );

	    // update the CSW structure 
	    //=========================
	    if ( mode_sense->allocation_length < sizeof(MODE_SENSE_PARAMETER) )
	    {
	      	msd_prepare_CSW( chan, GOOD_STATUS, 0);		// ?????			
	    }
	    else
	    {
	      	msd_prepare_CSW( chan, GOOD_STATUS, mode_sense->allocation_length-sizeof(MODE_SENSE_PARAMETER) );		// ?????
	    }
	    
  	}
  	else
  	{
    	usb_msd_send( chan, (char *)&sense.head, sizeof(USBMS_MODE_PARAMETER_HEAD) );
    	
    	// We return an empty structure 
    	// Update the CSW 
    	//=============================
    	msd_prepare_CSW( chan, GOOD_STATUS, mode_sense->allocation_length);		// ?????
    	
    	// update the state 
    	//=================
  //  	usbms_context.state = USB_SEND_NO_FLASH;
  	}


  return ENOERR;
}

//=============================================================================
// Function:        void msd_reset_sense_data(void)
//
// Overview:        This routine resets the Sense Data, initializing the
//                  structure REQ_SENSE_RESP gblSenseData.
//                  
//=============================================================================
static void msd_reset_sense_data( MSD_CHAN *chan ) 
{
	DUMP_MSG(5, "msd_reset_sense_data()\r\n");

	memset( chan->sense_data, 0, sizeof(chan->sense_data) );
	
	chan->sense_data[0] = S_CURRENT;
	chan->sense_data[7] = 10;				// n-7 (n=17 (0..17))
}


//================================================================================
// 					cdb_medium_removal()
//					--------------------
//
// Overview:        This function prepares response to Prevent Allow
//                  Medium Removal Command No data response is expect only 
//                  a CSW with command execution status is expected
//                  Since we cannot control the removal of media,
//                  we respond by a Success CSW
//================================================================================
static void cdb_medium_removal( MSD_CHAN *chan )
{
	DUMP_MSG(5, "cdb_medium_removal()\r\n");

	if( SDC_DetectCard() ) 
	{
		usb_msd_send_status( chan, GOOD_STATUS, 0 );			
	} 
	else 
	{
		SENSE_SET_SENSE_KEY( chan->sense_data, S_NOT_READY );
		SENSE_SET_ASC( chan->sense_data, ASC_MEDIUM_NOT_PRESENT );
		SENSE_SET_ASCQ( chan->sense_data, ASCQ_MEDIUM_NOT_PRESENT );

		usb_msd_send_status( chan, COMMAND_FAILED, 0 );	
	}	
}

//================================================================================
// Function:        void cdb_verify()
//
// Overview:        This function prepares response to Verify Command
//		            No data response is expected, we reply by a success CSW 
//		            The command is not being processed in this version of code
//
// Note:            None
//================================================================================
static void cdb_verify( MSD_CHAN *chan )
{
cyg_uint32 addr;
int len;

	addr = VALUE_32( chan->msd_cbw_copy.CBWCB[5], chan->msd_cbw_copy.CBWCB[4], chan->msd_cbw_copy.CBWCB[3], chan->msd_cbw_copy.CBWCB[2] );
	len  = (chan->msd_cbw_copy.CBWCB[7] << 8) + chan->msd_cbw_copy.CBWCB[8];
    
	DUMP_MSG(5, "cdb_verify( addr = %d, len = %d )\r\n", addr, len );
	
	usb_msd_send_status( chan, GOOD_STATUS, 0 );	
}

//================================================================================
// 					cdb_stop_start()
//					----------------
//
// Overview:        This function prepares response to Start Stop Unit Command
//                  No data response is expected, we reply by a success CSW 
//                  The command is not being processed in this version of code
//================================================================================
static void cdb_stop_start( MSD_CHAN *chan )
{
	DUMP_MSG(5, "cdb_stop_start()\r\n");
	
	usb_msd_send_status( chan, GOOD_STATUS, 0 );	
}	



//================================================================================
//					cdb_read_10()
//					-------------
//	Parameter:
//		chan ->
//
// 	Return: none
//
//	Description:
//		Read date from MMC/SD and send it to the HOST
//================================================================================
static void cdb_read_10( MSD_CHAN *chan )
{
int stat;
	
	DUMP_MSG(5, "cdb_read_10( addr = %d, len = %d )\r\n", chan->addr, chan->len );

	if ( chan->len == 0 )
	{
		usb_msd_send_status( chan, GOOD_STATUS, 0 );	
		return;
	}
		
	// read one page (512, 1024 or 2048 bytes)
	//========================================
	stat = SDC_SectorRead( chan, chan->addr, chan->msd_buffer );		
	if ( stat )
	{
		usb_msd_send_status( chan, COMMAND_FAILED, 0 );
		return;
	}

	// send requested data
	//====================
	usb_msd_send( chan, chan->msd_buffer, chan->mmc_page_size );

	chan->addr++;
	chan->len--;		

}


//================================================================================
//					cdb_write_10()
//					--------------
//	Parameter:
//		chan ->
//
// 	Return: none
//
//	Description:
//		Write data to the MMC/SD card
//================================================================================
static void cdb_write_10( MSD_CHAN *chan )
{
int stat;

	DUMP_MSG(5, "cdb_write_10( addr = %d, len = %d )\r\n", chan->addr, chan->len );

	stat = SDC_SectorWrite( chan, chan->addr, chan->msd_buffer);	
	if ( stat )
	{
//		usb_msd_send_status( chan, COMMAND_FAILED, 0 );
		return;
	}

	chan->addr++;
	chan->len--;		

}



static bool SDC_DetectCard(void)
{
    return true;
}

#ifdef MSD_DUMMY

cyg_uint8 SDC_Buffer[24*1024*1024];


static int SDC_SectorRead( MSD_CHAN *chan, cyg_uint32 sector_addr, cyg_uint8 *buffer )
{
int i;

	DUMP_MSG(5, "SDC_SectorRead(sect_addr = %d)\r\n", sector_addr);
    for ( i=0; i < 512;i++ )
        buffer[i]=SDC_Buffer[(sector_addr*512) +i];

    return ENOERR;
}

static int SDC_SectorWrite( MSD_CHAN *chan, cyg_uint32 sector_addr, cyg_uint8* buffer )
{
SDC_Error status=sdcValid;
int i;
    
	DUMP_MSG(5, "SDC_SectorWrite(sect_addr = %d)\r\n", sector_addr);
    if ( sector_addr >= sizeof(SDC_Buffer)/512 )
        return(status);
    
    for ( i=0; i < 512; i++ )
        SDC_Buffer[(sector_addr*512) +i]=buffer[i];    
       
    return(status);
}


#else

// ===============================================================
static int SDC_SectorRead( MSD_CHAN *chan, cyg_uint32 sector_addr, cyg_uint8 *buffer )
{
SDC_Error status = sdcValid;
int len, stat;

	DUMP_MSG(5, "SDC_SectorRead(sect_addr = %d)\r\n", sector_addr);

	len = 1;
	stat = cyg_io_bread(chan->mmc_handle, buffer, &len, sector_addr );
	if ( stat ) 
	{
		DUMP_MSG(0, "-- error: cyg_io_bread( sect_addr = %d ) returns %d\r\n", sector_addr, stat ); 
		status = sdcCardTimeout;
	}

#if (DUMP_LEVEL > 1) 
	if ( sector_addr == 0 )
		dump_buff( buffer, 512 );	
#endif		

    return(status);
}

static int SDC_SectorWrite( MSD_CHAN *chan, cyg_uint32 sector_addr, cyg_uint8* buffer )
{
SDC_Error status = sdcValid;
int len, stat;

	DUMP_MSG(5, "SDC_SectorWrite(sect_addr = %d)\r\n", sector_addr);


#if (DUMP_LEVEL > 1) 

	if ( sector_addr == 0 )
	{
		DUMP_MSG(0, "----------------- SDC_SectorWrite(0)\r\n");
		dump_buff( buffer, chan->mmc_page_size );	
	}
#endif 
    
	len = 1;
	stat = cyg_io_bwrite(chan->mmc_handle, buffer, &len, sector_addr );
	if ( stat ) 
	{
		DUMP_MSG(0, "-- error: cyg_io_bwrite( sect_addr = %d ) returns %d\r\n", sector_addr, stat ); 
		status = sdcCardTimeout;
	}
       
    return(status);
}


#endif

// ===============================================================

#define	PRINTF		diag_printf
static int dump_partition_table( char *buff )
{
const static cyg_uint8 supported_part_types[] = { 1,2,5,6,7,0xB,0xC,0xE,0xF };
#define	NUM_SUPP_PTYPES		(sizeof(supported_part_types)/sizeof(supported_part_types[0]))
int hh, cc, ss, ii;
cyg_uint32 rr;

	for (ii=0; ii < NUM_SUPP_PTYPES; ii++ )    	
		if ( buff[446+4] == supported_part_types[ii] ) 
			break;
		
	if ( ii >= NUM_SUPP_PTYPES )
	{
		PRINTF("++ info: no valid partition type -> 0x%X\r\n", buff[446+4]);
		PRINTF("++ info: maybe there is no MBR\r\n");
		return ENOTSUP;	
	}
	
	
	PRINTF("Partition Table:\r\n");

	hh = buff[446+1];
	ss = buff[446+2] & 0x3F;
	cc = buff[446+3] + ((buff[446+2] >> 6) & 3);
	
	PRINTF("  current state  = 0x%X (0 - inactive, 0x80 - active)\r\n", buff[446] );
	PRINTF("  partition type = %d\r\n", buff[446+4] );
	PRINTF("  begin C/H/S    = %d/%d/%d\r\n", cc, hh, ss );

	hh = buff[446+5];
	ss = buff[446+6] & 0x3F;
	cc = buff[446+7] + ((buff[446+6] >> 6) & 3);

	PRINTF("  end   C/H/S    = %d/%d/%d\r\n", cc, hh, ss );
	
	rr = buff[446+8] + (buff[446+9] << 8) + (buff[446+0xA] << 16) + (buff[446+0xB] << 24);
	
	PRINTF("  first sector   = %d\r\n", rr );

	rr = buff[446+0xC] + (buff[446+0xD] << 8) + (buff[446+0xE] << 16) + (buff[446+0xF] << 24);

	PRINTF("  num. sectors   = %d\r\n\n", rr );

	return ENOERR;	
}

//=============================================================================
//							usb_msd_init()
//							--------------
//
//=============================================================================
int usb_msd_init(void)
{
	memset( Msd_chans, 0, sizeof(Msd_chans) );
	return ENOERR;	
}

static int test_correct_mbr( MSD_CHAN *chan )
{
int len, stat, nn;

	// read PAGE 0
	//============
	stat = SDC_SectorRead( chan, 0, chan->msd_buffer );
	if ( stat ) return EINVAL;

	if ( (chan->msd_buffer[0x1FE] == 0x55) && (chan->msd_buffer[0x1FF] == 0xAA) )
	{
		dump_partition_table( chan->msd_buffer );
		return ENOERR;
	}

	memset( chan->msd_buffer, 0, MSD_MAX_MMC_PAGE );
	
	diag_printf("-- error: PAGE 0 is not valid\r\n");

	// partition table
	//================
    chan->msd_buffer[446+0] = 0x00;	// 0 - Current State of Partition (00h=Inactive, 80h=Active)
    chan->msd_buffer[446+1] = 0x0;	// 1 - Beginning of Partition - Head
    chan->msd_buffer[446+2] = 0x0;	// 2 - Beginning of Partition - Cylinder/Sector (See Below) LSB
    chan->msd_buffer[446+3] = 0x0;	// 3 - Beginning of Partition - Cylinder/Sector (See Below) MSB
    chan->msd_buffer[446+4] = 0x06;	// 4 - partition type
    chan->msd_buffer[446+5] = 0;	// 5 - End of Partition - Head
    chan->msd_buffer[446+6] = 0;	// 6 - End of Partition - Cylinder/Sector
    chan->msd_buffer[446+7] = 0;	// 7 - End of Partition - Cylinder/Sector
    
    // part->start
    // Number of Sectors Between the MBR and the First Sector in the Partition
    //=========================================================================
    nn = 64;
    chan->msd_buffer[446+0x8] = (cyg_uint8)nn;
    chan->msd_buffer[446+0x9] = (cyg_uint8)(nn >> 8);
    chan->msd_buffer[446+0xA] = (cyg_uint8)(nn >> 16);
    chan->msd_buffer[446+0xB] = (cyg_uint8)(nn >> 24);
    
    // part->size
    // Number of Sectors in the Partition
    //====================================
    len = chan->mmc_num_pages - nn;
    chan->msd_buffer[446+0xC] = (cyg_uint8)len;
    chan->msd_buffer[446+0xD] = (cyg_uint8)(len >> 8);
    chan->msd_buffer[446+0xE] = (cyg_uint8)(len >> 16);
    chan->msd_buffer[446+0xF] = (cyg_uint8)(len >> 24);
    
    chan->msd_buffer[0x1FE] = 0x55;	
    chan->msd_buffer[0x1FF] = 0xAA;

	diag_printf("++ info:  PAGE 0 corrected\r\n");


	stat = SDC_SectorWrite( chan, 0, chan->msd_buffer );
	if ( stat ) return EINVAL;
	
	return ENOERR;
}


//=============================================================================
//							usb_msd_open()
//							--------------
//	Parameter:
//		conf  ->
//		token ->
//
//	Return:
//		ENOERR
//
//	Description:
//		
//=============================================================================
int usb_msd_open( USB_MSD_CONF *conf_data, int *token )
{
MSD_CHAN *chan = &Msd_chans[0];
USB_MSD_CONF *conf;
int ii;

	*token = 0;

	for (ii=0; ii < USB_MSD_MAX_DEVICES; ii++, chan++ )
		if ( chan->flag_busy == false )
			break;
			
	if ( ii>= USB_MSD_MAX_DEVICES )
		return EINVAL;		

	conf = &chan->conf;
	chan->msd_buffer = (cyg_uint8 *)chan->msd_buffer_mem;
	chan->msd_buffer += SDRAM_NON_CACHE_OFFSET;

	// store the configuaration parameter
	//===================================
	memcpy( conf, conf_data, sizeof(USB_MSD_CONF) );

#ifdef MSD_DUMMY
	
	chan->mmc_page_size = 512;
	chan->mmc_num_pages = (sizeof(SDC_Buffer) / chan->mmc_page_size);
	memset( SDC_Buffer, 0, sizeof(SDC_Buffer) );

	diag_printf("++ info: using RAM emulation: page_size = %d, num_pages = %d\r\n", chan->mmc_page_size, chan->mmc_num_pages);
	
#else
	int stat, len;
	
	stat = cyg_io_lookup( conf->device, &chan->mmc_handle );	
	if ( stat ) return EINVAL;
		
	len = 4;
	cyg_io_get_config(chan->mmc_handle, CYG_IO_CONFIG_MMC_GET_RDBLOCK_SIZE, &chan->mmc_page_size, &len);

	len = 4;
	cyg_io_get_config(chan->mmc_handle, CYG_IO_CONFIG_MMC_GET_NUM_PAGES, &chan->mmc_num_pages, &len);

	// check for supported page size
	//==============================
	if ( chan->mmc_page_size > MSD_MAX_MMC_PAGE )
		return EINVAL;

	diag_printf("++info: MMC/SD card: page_size = %d, num_pages = %d\r\n", chan->mmc_page_size, chan->mmc_num_pages);

#endif

	
	// read and check page 0
	//======================
	test_correct_mbr( chan );
	
    
	conf->endp_ctrl->enumeration_data   = &usb_enum_data;
	conf->endp_ctrl->class_control_fn   = &MSD_ClassControlHandler;
	conf->endp_ctrl->class_control_data = (void*)chan;   		
	conf->endp_ctrl->state_change_fn    = &MSD_EndpointStateChange;
	conf->endp_ctrl->state_change_data  = (void*)chan;
    usbs_start( conf->endp_ctrl );

    // setup the buffer for bulk out transferes
    // ========================================
    conf->endp_out->buffer              = (unsigned char*)&chan->msd_cbw;
    conf->endp_out->buffer_size         = sizeof(chan->msd_cbw);
    conf->endp_out->complete_fn         = &MSD_BulkOutCallback;			// Deal with the packets HOST -> DEVICE
    conf->endp_out->complete_data       = (void*)chan;

    // setup the buffer for bulk in transferes
    // ========================================
    conf->endp_in->buffer            	= (unsigned char*)&chan->msd_buffer;
    conf->endp_in->buffer_size       	= sizeof( chan->msd_buffer_mem );
    conf->endp_in->complete_fn       	= &MSD_BulkInCallback;			// Deal with the packets DEVICE -> HOST
    conf->endp_in->complete_data     	= (void*)chan;

	chan->msd_state = USB_MSD_STATE_WAIT;    




	// create main thread
	//===================
  	cyg_thread_create(	10, 							// priority
  						usb_msd_task, 					// task function
  						(cyg_addrword_t) chan,			// data for task function	
		    			"USB Mass Storage Task", 		// name of task
		    			(void *) chan->task_stack, 		// pointer to stack
		    			sizeof(chan->task_stack),		// size of stack
		    			&chan->task_thread, 
		    			&chan->task_thread_s
		    			);

  	// start thread
  	//=============
  	cyg_thread_resume(chan->task_thread);


	chan->flag_busy = true;
	
	return ENOERR;
}




//=============================================================================
//							usb_msd_close()
//							---------------
//	Parameter:
//		token ->
//
//	Return:
//		ENOERR
//
//	Description:
//		
//=============================================================================
int usb_msd_close( int token )
{
MSD_CHAN *chan = (MSD_CHAN *)token;

	chan->flag_busy = false;	
	return ENOERR;
}

//=============================================================================
//							usb_msd_send()
//							--------------
//
//=============================================================================
static int usb_msd_send( MSD_CHAN *chan, void *buff, int len )
{
    chan->conf.endp_in->buffer            = (cyg_uint8 *)buff;
    chan->conf.endp_in->buffer_size       = len;
    usbs_start_tx(chan->conf.endp_in);

    return ENOERR;
}                        

//=============================================================================
//							usb_msd_send_status()
//							---------------------
//
//=============================================================================
static int usb_msd_send_status( MSD_CHAN *chan, int status, cyg_uint32 residue )
{
	DUMP_MSG(6, "usb_msd_send_status( status = %d, size = %d)\r\n", status, residue );
	
	msd_prepare_CSW( chan, status, residue );
	
	return usb_msd_send( chan, (cyg_uint8 *)&chan->msd_csw, CSW_SIZE );
}


//=============================================================================
//							usb_msd_task()
//							--------------
//
//=============================================================================
static void usb_msd_task(cyg_addrword_t data)
{
MSD_CHAN *chan = (MSD_CHAN *)data;
USBMS_BOT_CDB *_bot_cdb = NULL;

	while(1)
	{
		cyg_semaphore_wait(&chan->completion_wait);
        
        DUMP_MSG( 6, "%s\r\n", get_state_name( chan->msd_state ) );
        
        switch( chan->msd_state )
        {
            case USB_MSD_STATE_WAIT:

           	    _bot_cdb = (USBMS_BOT_CDB*)&chan->msd_cbw.CBWCB;
    			chan->curr_lun = chan->msd_cbw.bCBWLUN;

    			DUMP_MSG(6, "COMMAND = %s\r\n", get_ufi_name(_bot_cdb->generic_rbc.operation_code) );
            	memcpy( &chan->msd_cbw_copy, &chan->msd_cbw, sizeof(chan->msd_cbw) );	

                if ( MSD_IsValidCBW(chan) == false ) 
                	break;
                	
                if (MSD_IsMeaningfulCBW(chan) == false ) 
                	break;
                	
			    
                switch( _bot_cdb->generic_rbc.operation_code )
                {
                	case SCSI_INQUIRY:
                    	cdb_inquiry(chan);
                		chan->msd_state = USB_MSD_STATE_DATA_IN;
                        break;
                        
                    case SCSI_READ_CAPACITY:
                    	cdb_read_capacity( chan );	
				    	chan->msd_state = USB_MSD_STATE_DATA_IN;
					    break;          
					    
                    case SCSI_READ_FORMAT_CAPACITY:
                        cdb_read_format_capacity(chan);
                		chan->msd_state = USB_MSD_STATE_DATA_IN;
                        break;

					case SCSI_MODE_SELECT:
						cdb_mode_select(chan);
                		chan->msd_state = USB_MSD_STATE_DATA_IN;
						break;
                        
                    case SCSI_MODE_SENSE:
                    	cdb_mode_sense(chan);
					    chan->msd_state = USB_MSD_STATE_DATA_IN;
                	    break;
                	    
                    case SCSI_REQUEST_SENSE:
                	   	cdb_request_sense(chan);
					    chan->msd_state = USB_MSD_STATE_DATA_IN;
                        break;

                	case SCSI_TEST_UNIT_READY:
                	    cdb_test_unit_ready(chan);
                		chan->msd_state = USB_MSD_STATE_DATA_IN_END;
                        break;
                        
                	case SCSI_VERIFY:
                	   	cdb_verify(chan);
                		chan->msd_state = USB_MSD_STATE_DATA_IN_END;
                        break;
                        
                	case SCSI_PREVENT_ALLOW_MEDIUM_REMOVAL:
                	   	cdb_medium_removal(chan);
                		chan->msd_state = USB_MSD_STATE_DATA_IN_END;
                        break;
                        
                	case SCSI_STOP_START:
                	   	cdb_stop_start(chan);
                		chan->msd_state = USB_MSD_STATE_DATA_IN_END;
                        break; 

                  
                	case SCSI_READ_10:
                	{
                		USBMS_READ_10 *req = (USBMS_READ_10 *)chan->msd_cbw_copy.CBWCB;

						chan->addr = VALUE_32( req->block_address_0, req->block_address_1, req->block_address_2, req->block_address_3  );
						chan->len  = (req->xfer_len_1 << 8) + req->xfer_len_0;
                		
                		cdb_read_10( chan );	
                		chan->msd_state = USB_MSD_STATE_DATA_IN;
                		break;
                	}
                		
                	case SCSI_WRITE_10:
                	{
                		USBMS_WRITE_10 *req = (USBMS_WRITE_10 *)chan->msd_cbw_copy.CBWCB;

						chan->addr = VALUE_32( req->block_address_0, req->block_address_1, req->block_address_2, req->block_address_3  );
						chan->len  = (req->xfer_len_1 << 8) + req->xfer_len_0;

						if ( chan->len > 0 )
						{
                            chan->conf.endp_out->buffer      = chan->msd_buffer;
                            chan->conf.endp_out->buffer_size = chan->mmc_page_size;
                            usbs_start_rx(chan->conf.endp_out);
        				    chan->msd_state = USB_MSD_STATE_DATA_OUT;
						}
                		
                		break;
                	}
                		
                	default:
                		DUMP_MSG(0, "-- error: unsupported SCSI command: 0x%02X", chan->msd_cbw_copy.CBWCB[0] );
                    	msd_reset_sense_data(chan);
						SENSE_SET_SENSE_KEY( chan->sense_data, S_ILLEGAL_REQUEST );
						SENSE_SET_ASC( chan->sense_data, ASC_INVALID_COMMAND_OPCODE );
						SENSE_SET_ASCQ( chan->sense_data, ASCQ_INVALID_COMMAND_OPCODE );

						msd_prepare_CSW( chan, COMMAND_FAILED, 0 );
                		break;
                		
                } // end switch	
                        
                break;
				                
            case USB_MSD_STATE_DATA_IN_END:
                chan->conf.endp_out->buffer            = (unsigned char*)&chan->msd_cbw;
                chan->conf.endp_out->buffer_size       = sizeof(chan->msd_cbw);
                usbs_start_rx(chan->conf.endp_out);
                chan->msd_state = USB_MSD_STATE_WAIT;    
                break;
                    
            case USB_MSD_STATE_DATA_IN:

				switch( chan->msd_cbw_copy.CBWCB[0] )
				{
					case SCSI_INQUIRY:
					case SCSI_READ_FORMAT_CAPACITY:
					case SCSI_READ_CAPACITY:
					case SCSI_MODE_SENSE:
						usb_msd_send_status( chan, GOOD_STATUS, 0 );	
						chan->msd_state = USB_MSD_STATE_DATA_IN_END;
						break;
						
					case SCSI_READ_10:
	                	if ( chan->len > 0 )
	                		cdb_read_10( chan );	
	                	else
	                	{
	                		usb_msd_send_status( chan, GOOD_STATUS, 0 );
	                		chan->msd_state = USB_MSD_STATE_DATA_IN_END;
	                	}	
						break;						
				}
				
				break;
            	

            case USB_MSD_STATE_DATA_OUT:
                if ( chan->msd_cbw_copy.CBWCB[0] != SCSI_WRITE_10 )
                {	
					diag_printf("-- error: chan->msd_cbw_copy.CBWCB[0] != SCSI_WRITE_10\r\n");                	
                	break;
                }

            	cdb_write_10( chan );

            	if ( chan->len == 0 )
            	{
            		usb_msd_send_status( chan, GOOD_STATUS, 0 );
            		chan->msd_state = USB_MSD_STATE_DATA_IN_END;
            		break;
            	}

                chan->conf.endp_out->buffer            = chan->msd_buffer;
                chan->conf.endp_out->buffer_size       = chan->mmc_page_size;
                usbs_start_rx( chan->conf.endp_out );

				break;                	
                	
            default:
                chan->msd_state = USB_MSD_STATE_WAIT;
                break;
        }
	}
}


